import operator
import warnings
import weakref
from collections import (
    defaultdict,
    OrderedDict
)
from functools import wraps
from itertools import islice
from typing import (
    Any,
    Callable,
    Iterator,
    Optional,
    Set,
    Tuple,
    Union
)

import six

from joatmon.nn.utility import (
    _forward_unimplemented,
    EPOCH_DEPRECATION_WARNING,
    indent,
    legacy_get_string,
    required,
    typename
)

__all__ = ['Tensor', 'Module', 'Parameter', 'Optimizer', 'LRScheduler', 'Loss', 'ModuleAttributeException', 'Sequential']

warnings.filterwarnings(
    "once", "The .grad attribute of a Tensor that is not a leaf Tensor is being accessed. Its .grad "
            "attribute won't be populated during autograd.backward(). If you indeed want the gradient "
            "for a non-leaf Tensor, use .retain_grad() on the non-leaf Tensor. If you access the "
            "non-leaf Tensor by mistake, make sure you access the leaf Tensor instead.", UserWarning
)


class ModuleAttributeException(AttributeError):
    pass


class RemovableHandle(object):
    """A handle which provides the capability to remove a hook."""

    id: int
    next_id: int = 0

    def __init__(self, hooks_dict: Any) -> None:
        self.hooks_dict_ref = weakref.ref(hooks_dict)
        self.id = RemovableHandle.next_id
        RemovableHandle.next_id += 1

    def remove(self) -> None:
        hooks_dict = self.hooks_dict_ref()
        if hooks_dict is not None and self.id in hooks_dict:
            del hooks_dict[self.id]

    def __getstate__(self):
        return self.hooks_dict_ref(), self.id

    def __setstate__(self, state) -> None:
        if state[0] is None:
            # create a dead reference
            self.hooks_dict_ref = weakref.ref(OrderedDict())
        else:
            self.hooks_dict_ref = weakref.ref(state[0])
        self.id = state[1]
        RemovableHandle.next_id = max(RemovableHandle.next_id, self.id + 1)

    def __enter__(self) -> 'RemovableHandle':
        return self

    def __exit__(self, type: Any, value: Any, tb: Any) -> None:
        self.remove()


class Tensor:
    def __init__(self, data=None, requires_grad=None):
        self._data = data
        self._requires_grad = requires_grad
        self._grad_fn = None
        self._grad = None
        self._name = None
        self._backward_hooks = None
        self._device = 'cpu'

    def __repr__(self):
        return str(self)

    def __str__(self):
        name = f'name={self._name}' if self._name is not None else ''
        data = f'data={self._data}'
        require_grad = f'requires_grad={self._requires_grad}' if self._requires_grad is not None else ''
        grad_func = f'grad_fn=<{self._grad_fn.__name__}>' if self._grad_fn is not None else ''
        device = f'device={self.device}'

        str_list = [name, device, data, require_grad, grad_func]
        str_list = list(filter(lambda parameter: parameter != '', str_list))
        string = ', '.join(str_list)

        return f'Tensor({string})'

    def __getitem__(self, item):
        data = self.data[item]
        tensor = Tensor.from_array(data, requires_grad=self.requires_grad)
        tensor._grad_fn = self._grad_fn
        if self.device == 'gpu':
            tensor = tensor.gpu()
        return tensor

    def __setitem__(self, key, value):
        # self._grad_fn = value._grad_fn
        # self._requires_grad = value.requires_grad
        self._data[key] = value.data

    def __add__(self, other) -> 'Tensor':
        return self.add(other)

    def __sub__(self, other) -> 'Tensor':
        return self.sub(other)

    def __mul__(self, other) -> 'Tensor':
        return self.mul(other)

    def __truediv__(self, other) -> 'Tensor':
        return self.div(other)

    def __pow__(self, power, modulo=None) -> 'Tensor':
        return self.power(power)

    def __abs__(self) -> 'Tensor':
        return self.absolute()

    def chunk(self, chunks, dim=0):
        ...

    def view(self, size) -> 'Tensor':
        ...

    def index_select(self, dim, index) -> 'Tensor':
        ...

    def zero(self) -> 'Tensor':
        ...

    def one(self) -> 'Tensor':
        ...

    def fill(self, value) -> 'Tensor':
        ...

    def squeeze(self, axis=None) -> 'Tensor':
        ...

    def expand_dim(self, axis=None) -> 'Tensor':
        ...

    def transpose(self, axes) -> 'Tensor':
        ...

    def absolute(self) -> 'Tensor':
        ...

    def around(self) -> 'Tensor':
        ...

    def floor(self) -> 'Tensor':
        ...

    def ceil(self) -> 'Tensor':
        ...

    def clip(self, min_val, max_val) -> 'Tensor':
        ...

    def negative(self) -> 'Tensor':
        ...

    def summation(self) -> 'Tensor':
        ...

    def mean(self) -> 'Tensor':
        ...

    def std(self) -> 'Tensor':
        ...

    def var(self) -> 'Tensor':
        ...

    def add(self, other) -> 'Tensor':
        ...

    def sub(self, other) -> 'Tensor':
        ...

    def mul(self, other) -> 'Tensor':
        ...

    def div(self, other) -> 'Tensor':
        ...

    def power(self, p) -> 'Tensor':
        ...

    def clone(self) -> 'Tensor':
        ...

    def detach(self, inplace=False) -> 'Tensor':
        ...

    @staticmethod
    def from_array(data, requires_grad=False) -> 'Tensor':
        ...

    def to_array(self):
        ...

    def half(self) -> 'Tensor':
        ...

    def single(self) -> 'Tensor':
        ...

    def double(self) -> 'Tensor':
        ...

    def cpu(self) -> 'Tensor':
        ...

    def gpu(self) -> 'Tensor':
        ...

    def size(self, dim=None) -> Union[tuple, int]:
        if dim is None:
            return self._data.shape
        return self._data.shape[dim]

    def dim(self) -> int:
        return len(self._data.shape)

    @property
    def shape(self) -> tuple:
        return self._data.shape

    @property
    def ndim(self) -> int:
        return len(self._data.shape)

    @property
    def dtype(self):
        return self._data.dtype

    @property
    def device(self) -> str:
        return self._device

    @property
    def data(self):
        return self._data

    @property
    def is_leaf(self) -> bool:
        if not self._requires_grad:
            return True
        return self._grad_fn is None

    @property
    def grad(self) -> 'Tensor':
        if not self._can_read_grad():
            warnings.warn(
                "The .grad attribute of a Tensor that is not a leaf Tensor is being accessed. Its .grad "
                "attribute won't be populated during autograd.backward(). If you indeed want the gradient "
                "for a non-leaf Tensor, use .retain_grad() on the non-leaf Tensor. If you access the "
                "non-leaf Tensor by mistake, make sure you access the leaf Tensor instead.", UserWarning
            )

        return self._grad

    @grad.setter
    def grad(self, grad: 'Tensor'):
        self._grad = grad

    @property
    def requires_grad(self):
        return self._requires_grad

    @requires_grad.setter
    def requires_grad(self, requires_grad):
        if not requires_grad and not self.is_leaf:
            raise RuntimeError('you cannot set requires grad to False when the tensor is not a leaf tensor, you need to detach the tensor from graph')
        # if self.dtype != 'float32':
        #     raise RuntimeError('only float tensors can be required grad')
        self._requires_grad = requires_grad

    def retain_grad(self):
        if not self.requires_grad:
            raise RuntimeError("can't retain_grad on Tensor that has requires_grad=False")
        if self.is_leaf:  # no-op for leaves
            return
        if hasattr(self, 'retains_grad'):
            return

        import weakref
        weak_self = weakref.ref(self)

        def retain_grad_hook(grad):
            var = weak_self()
            if var is None:
                return
            if var._grad is None:
                var._grad = grad.clone()
                # if grad.is_sparse:
                #     var._grad = grad.clone()
                # else:
                #     var._grad = grad.clone(memory_format=torch.contiguous_format)
            else:
                var._grad = var._grad + grad

        self.register_hook(retain_grad_hook)

        self.retains_grad = True

    def register_hook(self, hook):
        if not self.requires_grad:
            raise RuntimeError("cannot register a hook on a tensor that doesn't require gradient")
        if self._backward_hooks is None:
            self._backward_hooks = OrderedDict()
            # if self._grad_fn is not None:
            #     self._grad_fn._register_hook_dict(self)
        handle = RemovableHandle(self._backward_hooks)
        self._backward_hooks[handle.id] = hook
        return handle

    def _can_read_grad(self):
        return not (self.requires_grad and not hasattr(self, "retains_grad") and not self.is_leaf and self._grad is None)

    def _can_write_grad(self):
        return not (self.requires_grad and not hasattr(self, "retains_grad") and not self.is_leaf)

    def forward(self):
        pass

    def backward(self, gradient=None):
        # if graph does not have any tensor that can have grad, RuntimeError: element 0 of tensors does not require grad and does not have a grad_fn

        import numpy as np
        if gradient is None:
            gradient = Tensor.from_array(np.array(1))

        if 'numpy' in type(gradient).__module__:
            gradient = Tensor.from_array(np.array(np.copy(gradient)))

        if self.device == 'gpu':
            gradient.gpu()

        if self._grad_fn is not None:
            self._grad_fn(gradient=gradient)

        if self._backward_hooks is not None:
            for _backward_hook in self._backward_hooks.values():
                _backward_hook(gradient)


class Parameter(Tensor):
    def __init__(self, data=None, requires_grad=True):
        if data is None:
            data = Tensor()

        super(Parameter, self).__init__(data=data.data, requires_grad=requires_grad)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return 'Parameter containing:\n' + super(Parameter, self).__str__()


class Module:
    dump_patches: bool = False
    _version: int = 1
    training: bool

    def __init__(self):
        self.training = True
        self._parameters = OrderedDict()
        self._modules = OrderedDict()

    def register_parameter(self, name: str, param: Optional[Parameter]) -> None:
        if '_parameters' not in self.__dict__:
            raise AttributeError("cannot assign parameter before Module.__init__() call")
        elif not isinstance(name, six.string_types):
            raise TypeError("parameter name should be a string. Got {}".format(typename(name)))
        elif '.' in name:
            raise KeyError("parameter name can't contain \".\"")
        elif name == '':
            raise KeyError("parameter name can't be empty string \"\"")
        elif hasattr(self, name) and name not in self._parameters:
            raise KeyError("attribute '{}' already exists".format(name))

        if param is None:
            self._parameters[name] = None
        elif not isinstance(param, Parameter):
            raise TypeError("cannot assign '{}' object to parameter '{}' (Parameter or None required)".format(typename(param), name))
        elif not param.is_leaf:
            raise ValueError(
                "Cannot assign non-leaf Tensor to parameter '{0}'. Model parameters must be created explicitly. "
                "To express '{0}' as a function of another Tensor, compute the value in the forward() method.".format(name)
            )
        else:
            self._parameters[name] = param

    def add_module(self, name: str, module: Optional['Module']) -> None:
        if not isinstance(module, Module) and module is not None:
            raise TypeError("{} is not a Module subclass".format(typename(module)))
        elif not isinstance(name, six.string_types):
            raise TypeError("module name should be a string. Got {}".format(typename(name)))
        elif hasattr(self, name) and name not in self._modules:
            raise KeyError("attribute '{}' already exists".format(name))
        elif '.' in name:
            raise KeyError("module name can't contain \".\"")
        elif name == '':
            raise KeyError("module name can't be empty string \"\"")
        self._modules[name] = module

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __getattr__(self, name: str) -> Union[Parameter, 'Module']:
        if '_parameters' in self.__dict__:
            _parameters = self.__dict__['_parameters']
            if name in _parameters:
                return _parameters[name]
        if '_modules' in self.__dict__:
            modules = self.__dict__['_modules']
            if name in modules:
                return modules[name]
        raise ModuleAttributeException("'{}' object has no attribute '{}'".format(type(self).__name__, name))

    def __setattr__(self, name: str, value: Union[Parameter, 'Module']) -> None:
        def remove_from(*dicts_or_sets):
            for d in dicts_or_sets:
                if name in d:
                    if isinstance(d, dict):
                        del d[name]
                    else:
                        d.discard(name)

        params = self.__dict__.get('_parameters')
        if isinstance(value, Parameter):
            if params is None:
                raise AttributeError("cannot assign parameters before Module.__init__() call")
            remove_from(self.__dict__, self._modules)
            self.register_parameter(name, value)
        elif params is not None and name in params:
            if value is not None:
                raise TypeError("cannot assign '{}' as parameter '{}' (Parameter or None expected)".format(typename(value), name))
            self.register_parameter(name, value)
        else:
            modules = self.__dict__.get('_modules')
            if isinstance(value, Module):
                if modules is None:
                    raise AttributeError("cannot assign module before Module.__init__() call")
                remove_from(self.__dict__, self._parameters)
                modules[name] = value
            elif modules is not None and name in modules:
                if value is not None:
                    raise TypeError("cannot assign '{}' as child module '{}' (Module or None expected)".format(typename(value), name))
                modules[name] = value
            else:
                object.__setattr__(self, name, value)

    forward: Callable[..., Any] = _forward_unimplemented

    def _call_impl(self, *inp, **kwargs):
        return self.forward(*inp, **kwargs)

    __call__: Callable[..., Any] = _call_impl

    def __delattr__(self, name):
        if name in self._parameters:
            del self._parameters[name]
        elif name in self._modules:
            del self._modules[name]
        else:
            object.__delattr__(self, name)

    def _named_members(self, get_members_fn, prefix='', recurse=True):
        memo = set()
        modules = self.named_modules(prefix=prefix) if recurse else [(prefix, self)]
        for module_prefix, module in modules:
            members = get_members_fn(module)
            for k, v in members:
                if v is None or v in memo:
                    continue
                memo.add(v)
                name = module_prefix + ('.' if module_prefix else '') + k
                yield name, v

    def parameters(self, recurse: bool = True) -> Iterator[Parameter]:
        for name, param in self.named_parameters(recurse=recurse):
            yield param

    def named_parameters(self, prefix: str = '', recurse: bool = True) -> Iterator[Tuple[str, Parameter]]:
        gen = self._named_members(lambda module: module._parameters.items(), prefix=prefix, recurse=recurse)
        for elem in gen:
            yield elem

    def children(self) -> Iterator['Module']:
        for name, module in self.named_children():
            yield module

    def named_children(self) -> Iterator[Tuple[str, 'Module']]:
        memo = set()
        for name, module in self._modules.items():
            if module is not None and module not in memo:
                memo.add(module)
                yield name, module

    def modules(self) -> Iterator['Module']:
        for name, module in self.named_modules():
            yield module

    def named_modules(self, memo: Optional[Set['Module']] = None, prefix: str = ''):
        if memo is None:
            memo = set()
        if self not in memo:
            memo.add(self)
            yield prefix, self
            for name, module in self._modules.items():
                if module is None:
                    continue
                submodule_prefix = prefix + ('.' if prefix else '') + name
                for m in module.named_modules(memo, submodule_prefix):
                    yield m

    # load_state_dict, save_state_dict, state_dict

    def train(self):
        self.training = True
        for module in self.children():
            module.train()
        return self

    def eval(self):
        self.training = False
        for module in self.children():
            module.eval()
        return self

    def zero_grad(self, set_to_none: bool = False):
        for p in self.parameters():
            if p.grad is not None:
                if set_to_none:
                    p.grad = None
                else:
                    if p.grad._grad_fn is not None:
                        p.grad.detach(inplace=True)
                    else:
                        p.grad.requires_grad = False
                    p.grad.fill(0)

    def state_dict(self):
        pass

    def load_state_dict(self):
        pass

    def save_state_dict(self):
        pass

    def _apply(self):
        pass

    def apply(self, func: Callable[['Module'], None]):
        for module in self.children():
            module.apply(func)
        func(self)
        return self

    def half(self):
        for param in self.parameters():
            param.half()
        return self

    def single(self):
        for param in self.parameters():
            param.single()
        return self

    def double(self):
        for param in self.parameters():
            param.double()
        return self

    def cpu(self):
        for param in self.parameters():
            param.cpu()
        return self

    def gpu(self):
        for param in self.parameters():
            param.gpu()
        return self

    def _get_name(self):
        return self.__class__.__name__

    def extra_repr(self) -> str:
        r"""Set the extra representation of the module

        To print customized extra information, you should re-implement
        this method in your own modules. Both single-line and multi-line
        strings are acceptable.
        """
        return ''

    def __repr__(self):
        extra_lines = []
        extra_repr = self.extra_repr()
        if extra_repr:
            extra_lines = extra_repr.split('\n')
        child_lines = []
        for key, module in self._modules.items():
            mod_str = repr(module)
            mod_str = indent(mod_str, 2)
            child_lines.append('(' + key + '): ' + mod_str)
        lines = extra_lines + child_lines

        main_str = self._get_name() + '('
        if lines:
            if len(extra_lines) == 1 and not child_lines:
                main_str += extra_lines[0]
            else:
                main_str += '\n  ' + '\n  '.join(lines) + '\n'

        main_str += ')'
        return main_str

    def __dir__(self):
        module_attrs = dir(self.__class__)
        attrs = list(self.__dict__.keys())
        parameters = list(self._parameters.keys())
        modules = list(self._modules.keys())
        keys = module_attrs + attrs + parameters + modules

        keys = [key for key in keys if not key[0].isdigit()]

        return sorted(keys)


class Loss(Module):
    reduction: str

    def __init__(self, size_average=None, reduce=None, reduction: str = 'mean') -> None:
        super(Loss, self).__init__()
        if size_average is not None or reduce is not None:
            self.reduction = legacy_get_string(size_average, reduce)
        else:
            self.reduction = reduction


class Sequential(Module):
    def __init__(self, *args: Any):
        super(Sequential, self).__init__()
        if len(args) == 1 and isinstance(args[0], OrderedDict):
            for key, module in args[0].items():
                self.add_module(key, module)
        else:
            for idx, module in enumerate(args):
                self.add_module(str(idx), module)

    def _get_item_by_idx(self, iterator, idx):
        """Get the idx-th item of the iterator"""
        size = len(self)
        idx = operator.index(idx)
        if not -size <= idx < size:
            raise IndexError('index {} is out of range'.format(idx))
        idx %= size
        return next(islice(iterator, idx, None))

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            return self.__class__(OrderedDict(list(self._modules.items())[idx]))
        else:
            return self._get_item_by_idx(self._modules.values(), idx)

    def __setitem__(self, idx: int, module: Module) -> None:
        key = self._get_item_by_idx(self._modules.keys(), idx)
        return setattr(self, key, module)

    def __delitem__(self, idx: Union[slice, int]) -> None:
        if isinstance(idx, slice):
            for key in list(self._modules.keys())[idx]:
                delattr(self, key)
        else:
            key = self._get_item_by_idx(self._modules.keys(), idx)
            delattr(self, key)

    def __len__(self) -> int:
        return len(self._modules)

    def __dir__(self):
        keys = super(Sequential, self).__dir__()
        keys = [key for key in keys if not key.isdigit()]
        return keys

    def __iter__(self) -> Iterator[Module]:
        return iter(self._modules.values())

    def forward(self, inp):
        for module in self:
            inp = module(inp)
        return inp


class Optimizer(object):
    def __init__(self, params, defaults):
        self.defaults = defaults

        if isinstance(params, Parameter):
            raise TypeError("params argument given to the optimizer should be an iterable of Tensors or dicts, but got " + typename(params))

        self.state = defaultdict(dict)
        self.param_groups = []

        param_groups = list(params)
        if len(param_groups) == 0:
            raise ValueError("optimizer got an empty parameter list")
        if not isinstance(param_groups[0], dict):
            param_groups = [{'params': param_groups}]

        for param_group in param_groups:
            self.add_param_group(param_group)

    def __getstate__(self):
        return {
            'defaults': self.defaults,
            'state': self.state,
            'param_groups': self.param_groups,
        }

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __repr__(self):
        format_string = self.__class__.__name__ + ' ('
        for i, group in enumerate(self.param_groups):
            format_string += '\n'
            format_string += 'Parameter Group {0}\n'.format(i)
            for key in sorted(group.keys()):
                if key != 'params':
                    format_string += '    {0}: {1}\n'.format(key, group[key])
        format_string += ')'
        return format_string

    def zero_grad(self, set_to_none: bool = False):
        for group in self.param_groups:
            for p in group['params']:
                if p.grad is not None:
                    if set_to_none:
                        p.grad = None
                    else:
                        if p.grad._grad_fn is not None:
                            p.grad.detach(inplace=True)
                        else:
                            p.grad._requires_grad = False
                        p.grad.zero()

    def step(self):
        raise NotImplementedError

    def add_param_group(self, param_group):
        assert isinstance(param_group, dict), "param group must be a dict"

        params = param_group['params']
        if isinstance(params, Parameter):
            param_group['params'] = [params]
        elif isinstance(params, set):
            raise TypeError(
                'optimizer parameters need to be organized in ordered collections, but '
                'the ordering of tensors in sets will change between runs. Please use a list instead.'
            )
        else:
            param_group['params'] = list(params)

        for param in param_group['params']:
            if not isinstance(param, Parameter):
                raise TypeError("optimizer can only optimize Tensors, but one of the params is " + typename(param))
            if not param.is_leaf:
                raise ValueError("can't optimize a non-leaf Tensor")

        for name, default in self.defaults.items():
            if default is required and name not in param_group:
                raise ValueError("parameter group didn't specify a value of required optimization parameter " + name)
            else:
                param_group.setdefault(name, default)

        params = param_group['params']
        if len(params) != len(set(params)):
            warnings.warn(
                "optimizer contains a parameter group with duplicate parameters; "
                "in future, this will cause an error", stacklevel=3
            )

        param_set = set()
        for group in self.param_groups:
            param_set.update(set(group['params']))

        if not param_set.isdisjoint(set(param_group['params'])):
            raise ValueError("some parameters appear in more than one parameter group")

        self.param_groups.append(param_group)


class LRScheduler(object):
    def __init__(self, optimizer, last_epoch=-1, verbose=False):

        if not isinstance(optimizer, Optimizer):
            raise TypeError(
                '{} is not an Optimizer'.format(
                    type(optimizer).__name__
                )
            )
        self.optimizer = optimizer

        if last_epoch == -1:
            for group in optimizer.param_groups:
                group.setdefault('initial_lr', group['lr'])
        else:
            for i, group in enumerate(optimizer.param_groups):
                if 'initial_lr' not in group:
                    raise KeyError("param 'initial_lr' is not specified in param_groups[{}] when resuming an optimizer".format(i))
        self.base_lrs = list(map(lambda group: group['initial_lr'], optimizer.param_groups))
        self.last_epoch = last_epoch

        def with_counter(method):
            if getattr(method, '_with_counter', False):
                return method

            instance_ref = weakref.ref(method.__self__)
            func = method.__func__
            cls = instance_ref().__class__
            del method

            @wraps(func)
            def wrapper(*args, **kwargs):
                instance = instance_ref()
                instance._step_count += 1
                wrapped = func.__get__(instance, cls)
                return wrapped(*args, **kwargs)

            wrapper._with_counter = True
            return wrapper

        self.optimizer.step = with_counter(self.optimizer.step)
        self.optimizer._step_count = 0
        self._step_count = 0
        self.verbose = verbose

        self.step()

    def state_dict(self):
        return {key: value for key, value in self.__dict__.items() if key != 'optimizer'}

    def load_state_dict(self, state_dict):
        self.__dict__.update(state_dict)

    def get_last_lr(self):
        return self._last_lr

    def get_lr(self):
        # Compute learning rate using chainable form of the scheduler
        raise NotImplementedError

    def print_lr(self, is_verbose, group, lr, epoch=None):
        if is_verbose:
            if epoch is None:
                print('Adjusting learning rate of group {} to {:.4e}.'.format(group, lr))
            else:
                print('Epoch {:5d}: adjusting learning rate of group {} to {:.4e}.'.format(epoch, group, lr))

    def step(self, epoch=None):
        # Raise a warning if old pattern is detected
        if self._step_count == 1:
            if not hasattr(self.optimizer.step, "_with_counter"):
                warnings.warn(
                    "Seems like `optimizer.step()` has been overridden after learning rate scheduler "
                    "initialization. Please, make sure to call `optimizer.step()` before "
                    "`lr_scheduler.step()`.", UserWarning
                )

            # Just check if there were two first lr_scheduler.step() calls before optimizer.step()
            elif self.optimizer._step_count < 1:
                warnings.warn(
                    "Detected call of `lr_scheduler.step()` before `optimizer.step()`. "
                    "In PyTorch 1.1.0 and later, you should call them in the opposite order: "
                    "`optimizer.step()` before `lr_scheduler.step()`.  Failure to do this "
                    "will result in PyTorch skipping the first value of the learning rate schedule.", UserWarning
                )
        self._step_count += 1

        class _enable_get_lr_call:

            def __init__(self, o):
                self.o = o

            def __enter__(self):
                self.o._get_lr_called_within_step = True
                return self

            def __exit__(self, type, value, traceback):
                self.o._get_lr_called_within_step = False

        with _enable_get_lr_call(self):
            if epoch is None:
                self.last_epoch += 1
                values = self.get_lr()
            else:
                warnings.warn(EPOCH_DEPRECATION_WARNING, UserWarning)
                self.last_epoch = epoch
                if hasattr(self, "_get_closed_form_lr"):
                    values = self._get_closed_form_lr()
                else:
                    values = self.get_lr()

        for i, data in enumerate(zip(self.optimizer.param_groups, values)):
            param_group, lr = data
            param_group['lr'] = lr
            self.print_lr(self.verbose, i, lr, epoch)

        self._last_lr = [group['lr'] for group in self.optimizer.param_groups]



from joatmon import context
from joatmon.plugin.logger.core import LoggerPlugin


class DatabaseLogger(LoggerPlugin):
    def __init__(self, level: str, database: str, cls, language, ip):
        super(DatabaseLogger, self).__init__(level, language, ip)

        self.database = database
        self.cls = cls

    async def _write(self, log: dict):
        database = context.get_value(self.database)
        await database.insert(self.cls, log)



from __future__ import print_function

import os

from joatmon.assistant.task import BaseTask


class Task(BaseTask):
    def __init__(self, api, **kwargs):
        super(Task, self).__init__(api, **kwargs)

    @staticmethod
    def help():
        return {
            "name": "cd",
            "description": "a function for user to change the current directory to given path",
            "parameters": {
                "type": "object",
                "properties": {
                    "path": {
                        "type": "string",
                        "description": "the path that the user want to change to"
                    }
                },
                "required": ["path"]
            }
        }

    def run(self):
        path = self.kwargs.get('path', None) or self.api.input('what is path that you want to change')

        parent_os_path = self.kwargs.get('parent_os_path', '')
        os_path = self.kwargs.get('os_path', '')

        match path:
            case '.':
                ...
            case '..':
                if os_path != os.sep:
                    os_path = os.sep.join(os_path.split(os.sep)[:-1])
                    if os_path == '':
                        os_path = os.sep
            case _:
                os_path = os.path.join(os_path, path)

        self.api.os_path = os_path

        self.api.output(f'current path is: {os_path}')

        if not self.stop_event.is_set():
            self.stop_event.set()



import uuid
from datetime import datetime

import psycopg2

from joatmon.orm.constraint import UniqueConstraint
from joatmon.orm.meta import normalize_kwargs
from joatmon.orm.query import (
    Dialects
)
from joatmon.plugin.database.core import DatabasePlugin
from joatmon.utility import get_converter


class PostgreSQLDatabase(DatabasePlugin):
    DATABASES = set()
    CREATED_COLLECTIONS = set()
    UPDATED_COLLECTIONS = set()

    def __init__(self, host, port, user, password, database):
        self.connection = psycopg2.connect(
            database=database, user=user,
            password=password, host=host, port=port  # , async_=True
        )
        self.connection.autocommit = True

    async def _check_collection(self, collection):
        # for one time only need to check indexes, constraints, default values, table schema as well
        cursor = self.connection.cursor()

        cursor.execute(f'select * from information_schema.tables where table_name = \'{collection.__collection__}\'')
        return len(list(cursor.fetchall())) > 0

    async def _create_collection(self, collection):
        def get_type(dtype: type):
            type_mapper = {
                datetime: 'timestamp without time zone',
                int: 'integer',
                float: 'real',
                str: 'varchar',
                bool: 'boolean',
                uuid.UUID: 'uuid',
            }

            return type_mapper.get(dtype, None)

        fields = []
        for field_name, field in collection.fields(collection).items():
            fields.append(f'{field_name} {get_type(field.dtype)} {"" if field.nullable else "not null"} {"primary key" if field.primary else ""}')
        sql = f'create table {collection.__collection__} (\n' + ',\n'.join(fields) + '\n);'

        cursor = self.connection.cursor()

        cursor.execute(sql)

        index_names = set()
        for index_name, index in collection.constraints(collection).items():
            if ',' in index.field:
                index_fields = list(map(lambda x: x.strip(), index.field.split(',')))
            else:
                index_fields = [index.field]
            c = ", ".join(index_fields)
            if index_name in index_names:
                continue
            index_names.add(index_name)
            cursor.execute(f'create {"unique" if isinstance(index, UniqueConstraint) else ""} index {collection.__collection__}_{index_name} on {collection.__collection__} ({c})')

    async def _create_view(self, collection):
        cursor = self.connection.cursor()

        sql = f'drop VIEW if exists {collection.__collection__}'
        cursor.execute(sql)

        sql = f'CREATE OR REPLACE VIEW {collection.__collection__} AS {collection.query(collection).build(Dialects.POSTGRESQL)}'
        cursor.execute(sql)

    async def _ensure_collection(self, collection):
        if not await self._check_collection(collection):
            await self._create_collection(collection)

    async def create(self, document):
        ...

    async def alter(self, document):
        ...

    async def drop(self, document):
        cursor = self.connection.cursor()
        sql = f'drop table if exists {document.__metaclass__.__collection__} cascade'
        cursor.execute(sql)

    async def insert(self, document, *docs):
        cursor = self.connection.cursor()

        for doc in docs:
            if not document.__metaclass__.structured:
                raise ValueError(f'you have to use structured document')

            await self._ensure_collection(document.__metaclass__)

            def normalize(d):
                dictionary = d.validate()
                fields = document.__metaclass__.fields(document.__metaclass__)

                keys = []
                values = []
                for field_name, field in fields.items():
                    keys.append(field_name)

                    if dictionary[field_name] is None:
                        values.append('null')
                    elif field.dtype in (uuid.UUID, str, datetime):
                        values.append(f'\'{str(dictionary[field_name])}\'')
                    else:
                        values.append(str(dictionary[field_name]))

                return keys, values, dictionary

            if isinstance(doc, dict):
                k, v, di = normalize(document(**doc))
            elif isinstance(doc, document):
                k, v, di = normalize(doc)
            else:
                raise ValueError(f'cannot convert object type {type(doc)} to {document}')
            sql = f'insert into {document.__metaclass__.__collection__} ({", ".join(k)}) values ({", ".join(v)})'

            cursor.execute(sql)

    async def read(self, document, query):
        cursor = self.connection.cursor()

        await self._ensure_collection(document.__metaclass__)

        keys = list(document.__metaclass__.fields(document.__metaclass__).keys())

        sql = f'select {", ".join(keys)} from {document.__metaclass__.__collection__}'

        def normalize(d, kwargs):
            fields = d.__metaclass__.fields(d.__metaclass__)

            keys = []
            values = []
            for k, v in kwargs.items():
                keys.append(k)

                field = fields[k]

                field_value = get_converter(field.dtype)(kwargs[k])

                if field_value is None:
                    values.append('null')
                elif field.dtype in (uuid.UUID, str, datetime):
                    values.append(f'\'{str(field_value)}\'')
                else:
                    values.append(str(field_value))

            return keys, values

        normalized = normalize_kwargs(document.__metaclass__, **query)
        k, v = normalize(document, normalized)

        if len(query) > 0:
            sql += f' where {" and ".join([f"{_k}={_v}" if _v != "null" else f"{_k} is {_v}" for _k, _v in zip(k, v)])}'

        cursor.execute(sql)

        # collection = await self._get_collection(document.__metaclass__.__collection__)
        # result = collection.find(
        #     normalize_kwargs(document.__metaclass__, **kwargs), {'_id': 0}, session=self.session
        # )

        for doc in cursor.fetchall():
            yield document(**dict(zip(keys, doc)))

    async def update(self, document, query, update):
        cursor = self.connection.cursor()

        def normalize(d, kwargs):
            fields = d.__metaclass__.fields(d.__metaclass__)

            keys = []
            values = []
            for k, v in kwargs.items():
                keys.append(k)

                field = fields[k]

                field_value = get_converter(field.dtype)(kwargs[k])

                if field_value is None:
                    values.append('null')
                elif field.dtype in (uuid.UUID, str, datetime):
                    values.append(f'\'{str(field_value)}\'')
                else:
                    values.append(str(field_value))

            return keys, values

        k, v = normalize(document, update)

        sql = f'update {document.__metaclass__.__collection__} set {", ".join(f"{_k}={_v}" for _k, _v in zip(k, v))}'

        def normalize(d, kwargs):
            fields = d.__metaclass__.fields(d.__metaclass__)

            keys = []
            values = []
            for k, v in kwargs.items():
                keys.append(k)

                field = fields[k]

                field_value = get_converter(field.dtype)(kwargs[k])

                if field_value is None:
                    values.append('null')
                elif field.dtype in (uuid.UUID, str, datetime):
                    values.append(f'\'{str(field_value)}\'')
                else:
                    values.append(str(field_value))

            return keys, values

        normalized = normalize_kwargs(document.__metaclass__, **query)
        k, v = normalize(document, normalized)

        if len(query) > 0:
            sql += f' where {" and ".join([f"{_k}={_v}" if _v != "null" else f"{_k} is {_v}" for _k, _v in zip(k, v)])}'

        await self._ensure_collection(document.__metaclass__)
        cursor.execute(sql)

    async def delete(self, document, query):
        cursor = self.connection.cursor()

        sql = f'delete from {document.__metaclass__.__collection__}'

        def normalize(d, kwargs):
            fields = d.__metaclass__.fields(d.__metaclass__)

            keys = []
            values = []
            for k, v in kwargs.items():
                keys.append(k)

                field = fields[k]

                field_value = get_converter(field.dtype)(kwargs[k])

                if field_value is None:
                    values.append('null')
                elif field.dtype in (uuid.UUID, str, datetime):
                    values.append(f'\'{str(field_value)}\'')
                else:
                    values.append(str(field_value))

            return keys, values

        normalized = normalize_kwargs(document.__metaclass__, **query)
        k, v = normalize(document, normalized)

        if len(query) > 0:
            sql += f' where {" and ".join([f"{_k}={_v}" if _v != "null" else f"{_k} is {_v}" for _k, _v in zip(k, v)])}'

        await self._ensure_collection(document.__metaclass__)
        cursor.execute(sql)

    async def view(self, document, query):
        await self._create_view(document.__metaclass__)

        keys = list(document.__metaclass__.fields(document.__metaclass__).keys())

        sql = f'select {", ".join(keys)} from {document.__metaclass__.__collection__}'

        def normalize(d, kwargs):
            fields = d.__metaclass__.fields(d.__metaclass__)

            keys = []
            values = []
            for k, v in kwargs.items():
                keys.append(k)

                field = fields[k]

                field_value = get_converter(field.dtype)(kwargs[k])

                if field_value is None:
                    values.append('null')
                elif field.dtype in (uuid.UUID, str, datetime):
                    values.append(f'\'{str(field_value)}\'')
                else:
                    values.append(str(field_value))

            return keys, values

        normalized = normalize_kwargs(document.__metaclass__, **query)
        k, v = normalize(document, normalized)

        if len(query) > 0:
            sql += f' where {" and ".join([f"{_k}={_v}" if _v != "null" else f"{_k} is {_v}" for _k, _v in zip(k, v)])}'

        cursor = self.connection.cursor()
        cursor.execute(sql)

        # keys = list(document.__metaclass__.fields(document.__metaclass__).keys())
        keys = [desc[0] for desc in cursor.description]
        for doc in cursor.fetchall():
            yield document(**dict(zip(keys, doc)))

    async def start(self):
        self.connection.autocommit = False

    async def commit(self):
        self.connection.commit()

    async def abort(self):
        self.connection.rollback()

    async def end(self):
        self.connection.close()



from __future__ import print_function

import json

import requests

from joatmon.assistant.task import BaseTask


class Task(BaseTask):
    def __init__(self, api, **kwargs):
        super(Task, self).__init__(api, **kwargs)

    @staticmethod
    def help():
        return {
            "name": "exchange",
            "description": "a function for user to learn exchange rates compared to TRY",
            "parameters": {
                "type": "object",
                "properties": {
                },
                "required": []
            }
        }

    def run(self):
        key = json.loads(open('iva/iva.json', 'r').read())['config']['exchange']['key']

        data = []

        resp = requests.get('https://api.twelvedata.com/exchange_rate', params={'apikey': key, 'symbol': 'USD/TRY'})
        response = json.loads(resp.content.decode('utf-8'))
        data.append(f'{response["symbol"]} : {response["rate"]}')

        resp = requests.get('https://api.twelvedata.com/exchange_rate', params={'apikey': key, 'symbol': 'EUR/TRY'})
        response = json.loads(resp.content.decode('utf-8'))
        data.append(f'{response["symbol"]} : {response["rate"]}')

        self.api.output('\n'.join(data))

        if not self.stop_event.is_set():
            self.stop_event.set()



import asyncio
from datetime import datetime
from struct import unpack

from ..adapters import (
    ReaderAdapter,
    WriterAdapter
)
from ..codecs import (
    bytes_to_hex_str,
    bytes_to_int,
    decode_data_with_length,
    decode_packet_id,
    decode_string,
    encode_data_with_length,
    encode_string,
    int_to_bytes,
    read_or_raise
)
from ..errors import (
    CodecException,
    HZMQTTException,
    MQTTException,
    NoDataException
)
from ..utils import gen_client_id

RESERVED_0 = 0x00
CONNECT = 0x01
CONNACK = 0x02
PUBLISH = 0x03
PUBACK = 0x04
PUBREC = 0x05
PUBREL = 0x06
PUBCOMP = 0x07
SUBSCRIBE = 0x08
SUBACK = 0x09
UNSUBSCRIBE = 0x0a
UNSUBACK = 0x0b
PINGREQ = 0x0c
PINGRESP = 0x0d
DISCONNECT = 0x0e
RESERVED_15 = 0x0f

CONNECTION_ACCEPTED = 0x00
UNACCEPTABLE_PROTOCOL_VERSION = 0x01
IDENTIFIER_REJECTED = 0x02
SERVER_UNAVAILABLE = 0x03
BAD_USERNAME_PASSWORD = 0x04
NOT_AUTHORIZED = 0x05


class MQTTFixedHeader:
    __slots__ = ('packet_type', 'remaining_length', 'flags')

    def __init__(self, packet_type, flags=0, length=0):
        self.packet_type = packet_type
        self.remaining_length = length
        self.flags = flags

    def to_bytes(self):
        def encode_remaining_length(length: int):
            encoded = bytearray()
            while True:
                length_byte = length % 0x80
                length //= 0x80
                if length > 0:
                    length_byte |= 0x80
                encoded.append(length_byte)
                if length <= 0:
                    break
            return encoded

        out = bytearray()
        packet_type = 0
        try:
            packet_type = (self.packet_type << 4) | self.flags
            out.append(packet_type)
        except OverflowError:
            raise CodecException('packet_type encoding exceed 1 byte length: value=%d', packet_type)

        encoded_length = encode_remaining_length(self.remaining_length)
        out.extend(encoded_length)

        return out

    async def to_stream(self, writer: WriterAdapter):
        writer.write(self.to_bytes())

    @property
    def bytes_length(self):
        return len(self.to_bytes())

    @classmethod
    async def from_stream(cls, reader: ReaderAdapter):
        async def decode_remaining_length():
            multiplier = 1
            value = 0
            buffer = bytearray()
            while True:
                encoded_byte = await reader.read(1)
                int_byte = unpack('!B', encoded_byte)
                buffer.append(int_byte[0])
                value += (int_byte[0] & 0x7f) * multiplier
                if (int_byte[0] & 0x80) == 0:
                    break
                else:
                    multiplier *= 128
                    if multiplier > 128 * 128 * 128:
                        raise MQTTException("Invalid remaining length bytes:%s, packet_type=%d" % (bytes_to_hex_str(buffer), msg_type))
            return value

        try:
            byte1 = await read_or_raise(reader, 1)
            int1 = unpack('!B', byte1)
            msg_type = (int1[0] & 0xf0) >> 4
            flags = int1[0] & 0x0f
            remain_length = await decode_remaining_length()

            return cls(msg_type, flags, remain_length)
        except NoDataException:
            return None

    def __repr__(self):
        return type(self).__name__ + '(length={0}, flags={1})'. \
            format(self.remaining_length, hex(self.flags))


class MQTTVariableHeader:
    def __init__(self):
        pass

    async def to_stream(self, writer: asyncio.StreamWriter):
        writer.write(self.to_bytes())
        await writer.drain()

    def to_bytes(self) -> bytes:
        ...

    @property
    def bytes_length(self):
        return len(self.to_bytes())

    @classmethod
    async def from_stream(cls, reader: asyncio.StreamReader, fixed_header: MQTTFixedHeader):
        pass


class PacketIdVariableHeader(MQTTVariableHeader):
    __slots__ = ('packet_id',)

    def __init__(self, packet_id):
        super().__init__()
        self.packet_id = packet_id

    def to_bytes(self):
        out = b''
        out += int_to_bytes(self.packet_id, 2)
        return out

    @classmethod
    async def from_stream(cls, reader: ReaderAdapter, fixed_header: MQTTFixedHeader):
        packet_id = await decode_packet_id(reader)
        return cls(packet_id)

    def __repr__(self):
        return type(self).__name__ + '(packet_id={0})'.format(self.packet_id)


class MQTTPayload:
    def __init__(self):
        pass

    async def to_stream(self, writer: asyncio.StreamWriter):
        writer.write(self.to_bytes())
        await writer.drain()

    def to_bytes(self, fixed_header: MQTTFixedHeader, variable_header: MQTTVariableHeader):
        pass

    @classmethod
    async def from_stream(cls, reader: asyncio.StreamReader, fixed_header: MQTTFixedHeader, variable_header: MQTTVariableHeader):
        pass


class MQTTPacket:
    __slots__ = ('fixed_header', 'variable_header', 'payload', 'protocol_ts')

    FIXED_HEADER = MQTTFixedHeader
    VARIABLE_HEADER = None
    PAYLOAD = None

    def __init__(self, fixed: MQTTFixedHeader, variable_header: MQTTVariableHeader = None, payload: MQTTPayload = None):
        self.fixed_header = fixed
        self.variable_header = variable_header
        self.payload = payload
        self.protocol_ts = None

    async def to_stream(self, writer: asyncio.StreamWriter):
        writer.write(self.to_bytes())
        await writer.drain()
        self.protocol_ts = datetime.now()

    def to_bytes(self) -> bytes:
        if self.variable_header:
            variable_header_bytes = self.variable_header.to_bytes()
        else:
            variable_header_bytes = b''
        if self.payload:
            payload_bytes = self.payload.to_bytes(self.fixed_header, self.variable_header)
        else:
            payload_bytes = b''

        self.fixed_header.remaining_length = len(variable_header_bytes) + len(payload_bytes)
        fixed_header_bytes = self.fixed_header.to_bytes()

        return fixed_header_bytes + variable_header_bytes + payload_bytes

    @classmethod
    async def from_stream(cls, reader: ReaderAdapter, fixed_header=None, variable_header=None):
        if fixed_header is None:
            fixed_header = await cls.FIXED_HEADER.from_stream(reader)
        if cls.VARIABLE_HEADER:
            if variable_header is None:
                variable_header = await cls.VARIABLE_HEADER.from_stream(reader, fixed_header)
        else:
            variable_header = None
        if cls.PAYLOAD:
            payload = await cls.PAYLOAD.from_stream(reader, fixed_header, variable_header)
        else:
            payload = None

        if fixed_header and not variable_header and not payload:
            instance = cls(fixed_header)
        elif fixed_header and not payload:
            instance = cls(fixed_header, variable_header)
        else:
            instance = cls(fixed_header, variable_header, payload)
        instance.protocol_ts = datetime.now()
        return instance

    @property
    def bytes_length(self):
        return len(self.to_bytes())

    def __repr__(self):
        return type(self).__name__ + '(ts={0!s}, fixed={1!r}, variable={2!r}, payload={3!r})'.format(self.protocol_ts, self.fixed_header, self.variable_header, self.payload)


class ConnackVariableHeader(MQTTVariableHeader):
    __slots__ = ('session_parent', 'return_code')

    def __init__(self, session_parent=None, return_code=None):
        super().__init__()
        self.session_parent = session_parent
        self.return_code = return_code

    @classmethod
    async def from_stream(cls, reader: ReaderAdapter, fixed_header: MQTTFixedHeader):
        data = await read_or_raise(reader, 2)
        session_parent = data[0] & 0x01
        return_code = bytes_to_int(data[1])
        return cls(session_parent, return_code)

    def to_bytes(self):
        out = bytearray(2)
        if self.session_parent:
            out[0] = 1
        else:
            out[0] = 0
        out[1] = self.return_code
        return out

    def __repr__(self):
        return type(self).__name__ + '(session_parent={0}, return_code={1})'.format(hex(self.session_parent), hex(self.return_code))


class ConnackPacket(MQTTPacket):
    VARIABLE_HEADER = ConnackVariableHeader
    PAYLOAD = None

    @property
    def return_code(self):
        return self.variable_header.return_code

    @return_code.setter
    def return_code(self, return_code):
        self.variable_header.return_code = return_code

    @property
    def session_parent(self):
        return self.variable_header.session_parent

    @session_parent.setter
    def session_parent(self, session_parent):
        self.variable_header.session_parent = session_parent

    def __init__(self, fixed: MQTTFixedHeader = None, variable_header: ConnackVariableHeader = None, payload=None):
        if fixed is None:
            header = MQTTFixedHeader(CONNACK, 0x00)
        else:
            if fixed.packet_type is not CONNACK:
                raise HZMQTTException("Invalid fixed packet type %s for ConnackPacket init" % fixed.packet_type)
            header = fixed
        super().__init__(header)
        self.variable_header = variable_header
        self.payload = None

    @classmethod
    def build(cls, session_parent=None, return_code=None):
        v_header = ConnackVariableHeader(session_parent, return_code)
        packet = ConnackPacket(variable_header=v_header)
        return packet


class ConnectVariableHeader(MQTTVariableHeader):
    __slots__ = ('proto_name', 'proto_level', 'flags', 'keep_alive')

    USERNAME_FLAG = 0x80
    PASSWORD_FLAG = 0x40
    WILL_RETAIN_FLAG = 0x20
    WILL_FLAG = 0x04
    WILL_QOS_MASK = 0x18
    CLEAN_SESSION_FLAG = 0x02
    RESERVED_FLAG = 0x01

    def __init__(self, connect_flags=0x00, keep_alive=0, proto_name='MQTT', proto_level=0x04):
        super().__init__()
        self.proto_name = proto_name
        self.proto_level = proto_level
        self.flags = connect_flags
        self.keep_alive = keep_alive

    def __repr__(self):
        return "ConnectVariableHeader(proto_name={0}, proto_level={1}, flags={2}, keepalive={3})".format(self.proto_name, self.proto_level, hex(self.flags), self.keep_alive)

    def _set_flag(self, val, mask):
        if val:
            self.flags |= mask
        else:
            self.flags &= ~mask

    def _get_flag(self, mask):
        if self.flags & mask:
            return True
        else:
            return False

    @property
    def username_flag(self) -> bool:
        return self._get_flag(self.USERNAME_FLAG)

    @username_flag.setter
    def username_flag(self, val: bool):
        self._set_flag(val, self.USERNAME_FLAG)

    @property
    def password_flag(self) -> bool:
        return self._get_flag(self.PASSWORD_FLAG)

    @password_flag.setter
    def password_flag(self, val: bool):
        self._set_flag(val, self.PASSWORD_FLAG)

    @property
    def will_retain_flag(self) -> bool:
        return self._get_flag(self.WILL_RETAIN_FLAG)

    @will_retain_flag.setter
    def will_retain_flag(self, val: bool):
        self._set_flag(val, self.WILL_RETAIN_FLAG)

    @property
    def will_flag(self) -> bool:
        return self._get_flag(self.WILL_FLAG)

    @will_flag.setter
    def will_flag(self, val: bool):
        self._set_flag(val, self.WILL_FLAG)

    @property
    def clean_session_flag(self) -> bool:
        return self._get_flag(self.CLEAN_SESSION_FLAG)

    @clean_session_flag.setter
    def clean_session_flag(self, val: bool):
        self._set_flag(val, self.CLEAN_SESSION_FLAG)

    @property
    def reserved_flag(self) -> bool:
        return self._get_flag(self.RESERVED_FLAG)

    @property
    def will_qos(self):
        return (self.flags & 0x18) >> 3

    @will_qos.setter
    def will_qos(self, val: int):
        self.flags &= 0xe7  # Reset QOS flags
        self.flags |= (val << 3)

    @classmethod
    async def from_stream(cls, reader: ReaderAdapter, fixed_header: MQTTFixedHeader):
        protocol_name = await decode_string(reader)

        protocol_level_byte = await read_or_raise(reader, 1)
        protocol_level = bytes_to_int(protocol_level_byte)

        flags_byte = await read_or_raise(reader, 1)
        flags = bytes_to_int(flags_byte)

        keep_alive_byte = await read_or_raise(reader, 2)
        keep_alive = bytes_to_int(keep_alive_byte)

        return cls(flags, keep_alive, protocol_name, protocol_level)

    def to_bytes(self):
        out = bytearray()
        out.extend(encode_string(self.proto_name))
        out.append(self.proto_level)
        out.append(self.flags)
        out.extend(int_to_bytes(self.keep_alive, 2))
        return out


class ConnectPayload(MQTTPayload):
    __slots__ = ('client_id', 'will_topic', 'will_message', 'username', 'password', 'client_id_is_random')

    def __init__(self, client_id=None, will_topic=None, will_message=None, username=None, password=None):
        super().__init__()
        self.client_id_is_random = False
        self.client_id = client_id
        self.will_topic = will_topic
        self.will_message = will_message
        self.username = username
        self.password = password

    def __repr__(self):
        return "ConnectVariableHeader(client_id={0}, will_topic={1}, will_message={2}, username={3}, password={4})".format(
            self.client_id, self.will_topic, self.will_message, self.username, self.password
        )

    @classmethod
    async def from_stream(cls, reader: ReaderAdapter, fixed_header: MQTTFixedHeader, variable_header: ConnectVariableHeader):
        payload = cls()
        try:
            payload.client_id = await decode_string(reader)
        except NoDataException:
            payload.client_id = None

        if payload.client_id is None or payload.client_id == "":
            payload.client_id = gen_client_id()
            payload.client_id_is_random = True

        if variable_header.will_flag:
            try:
                payload.will_topic = await decode_string(reader)
                payload.will_message = await decode_data_with_length(reader)
            except NoDataException:
                payload.will_topic = None
                payload.will_message = None

        if variable_header.username_flag:
            try:
                payload.username = await decode_string(reader)
            except NoDataException:
                payload.username = None

        if variable_header.password_flag:
            try:
                payload.password = await decode_string(reader)
            except NoDataException:
                payload.password = None

        return payload

    def to_bytes(self, fixed_header: MQTTFixedHeader, variable_header: ConnectVariableHeader):
        out = bytearray()
        out.extend(encode_string(self.client_id))
        if variable_header.will_flag:
            out.extend(encode_string(self.will_topic))
            out.extend(encode_data_with_length(self.will_message))
        if variable_header.username_flag:
            out.extend(encode_string(self.username))
        if variable_header.password_flag:
            out.extend(encode_string(self.password))
        return out


class ConnectPacket(MQTTPacket):
    VARIABLE_HEADER = ConnectVariableHeader
    PAYLOAD = ConnectPayload

    @property
    def proto_name(self):
        return self.variable_header.proto_name

    @proto_name.setter
    def proto_name(self, name: str):
        self.variable_header.proto_name = name

    @property
    def proto_level(self):
        return self.variable_header.proto_level

    @proto_level.setter
    def proto_level(self, level):
        self.variable_header.proto_level = level

    @property
    def username_flag(self):
        return self.variable_header.username_flag

    @username_flag.setter
    def username_flag(self, flag):
        self.variable_header.username_flag = flag

    @property
    def password_flag(self):
        return self.variable_header.password_flag

    @password_flag.setter
    def password_flag(self, flag):
        self.variable_header.password_flag = flag

    @property
    def clean_session_flag(self):
        return self.variable_header.clean_session_flag

    @clean_session_flag.setter
    def clean_session_flag(self, flag):
        self.variable_header.clean_session_flag = flag

    @property
    def will_retain_flag(self):
        return self.variable_header.will_retain_flag

    @will_retain_flag.setter
    def will_retain_flag(self, flag):
        self.variable_header.will_retain_flag = flag

    @property
    def will_qos(self):
        return self.variable_header.will_qos

    @will_qos.setter
    def will_qos(self, flag):
        self.variable_header.will_qos = flag

    @property
    def will_flag(self):
        return self.variable_header.will_flag

    @will_flag.setter
    def will_flag(self, flag):
        self.variable_header.will_flag = flag

    @property
    def reserved_flag(self):
        return self.variable_header.reserved_flag

    @reserved_flag.setter
    def reserved_flag(self, flag):
        self.variable_header.reserved_flag = flag

    @property
    def client_id(self):
        return self.payload.client_id

    @client_id.setter
    def client_id(self, client_id):
        self.payload.client_id = client_id

    @property
    def client_id_is_random(self) -> bool:
        return self.payload.client_id_is_random

    @client_id_is_random.setter
    def client_id_is_random(self, client_id_is_random: bool):
        self.payload.client_id_is_random = client_id_is_random

    @property
    def will_topic(self):
        return self.payload.will_topic

    @will_topic.setter
    def will_topic(self, will_topic):
        self.payload.will_topic = will_topic

    @property
    def will_message(self):
        return self.payload.will_message

    @will_message.setter
    def will_message(self, will_message):
        self.payload.will_message = will_message

    @property
    def username(self):
        return self.payload.username

    @username.setter
    def username(self, username):
        self.payload.username = username

    @property
    def password(self):
        return self.payload.password

    @password.setter
    def password(self, password):
        self.payload.password = password

    @property
    def keep_alive(self):
        return self.variable_header.keep_alive

    @keep_alive.setter
    def keep_alive(self, keep_alive):
        self.variable_header.keep_alive = keep_alive

    def __init__(self, fixed: MQTTFixedHeader = None, vh: ConnectVariableHeader = None, payload: ConnectPayload = None):
        if fixed is None:
            header = MQTTFixedHeader(CONNECT, 0x00)
        else:
            if fixed.packet_type is not CONNECT:
                raise HZMQTTException("Invalid fixed packet type %s for ConnectPacket init" % fixed.packet_type)
            header = fixed
        super().__init__(header)
        self.variable_header = vh
        self.payload = payload


class DisconnectPacket(MQTTPacket):
    VARIABLE_HEADER = None
    PAYLOAD = None

    def __init__(self, fixed: MQTTFixedHeader = None):
        if fixed is None:
            header = MQTTFixedHeader(DISCONNECT, 0x00)
        else:
            if fixed.packet_type is not DISCONNECT:
                raise HZMQTTException("Invalid fixed packet type %s for DisconnectPacket init" % fixed.packet_type)
            header = fixed
        super().__init__(header)
        self.variable_header = None
        self.payload = None


class PingReqPacket(MQTTPacket):
    VARIABLE_HEADER = None
    PAYLOAD = None

    def __init__(self, fixed: MQTTFixedHeader = None):
        if fixed is None:
            header = MQTTFixedHeader(PINGREQ, 0x00)
        else:
            if fixed.packet_type is not PINGREQ:
                raise HZMQTTException("Invalid fixed packet type %s for PingReqPacket init" % fixed.packet_type)
            header = fixed
        super().__init__(header)
        self.variable_header = None
        self.payload = None


class PingRespPacket(MQTTPacket):
    VARIABLE_HEADER = None
    PAYLOAD = None

    def __init__(self, fixed: MQTTFixedHeader = None):
        if fixed is None:
            header = MQTTFixedHeader(PINGRESP, 0x00)
        else:
            if fixed.packet_type is not PINGRESP:
                raise HZMQTTException("Invalid fixed packet type %s for PingRespPacket init" % fixed.packet_type)
            header = fixed
        super().__init__(header)
        self.variable_header = None
        self.payload = None

    @classmethod
    def build(cls):
        return cls()


class PubackPacket(MQTTPacket):
    VARIABLE_HEADER = PacketIdVariableHeader
    PAYLOAD = None

    @property
    def packet_id(self):
        return self.variable_header.packet_id

    @packet_id.setter
    def packet_id(self, val: int):
        self.variable_header.packet_id = val

    def __init__(self, fixed: MQTTFixedHeader = None, variable_header: PacketIdVariableHeader = None):
        if fixed is None:
            header = MQTTFixedHeader(PUBACK, 0x00)
        else:
            if fixed.packet_type is not PUBACK:
                raise HZMQTTException("Invalid fixed packet type %s for PubackPacket init" % fixed.packet_type)
            header = fixed
        super().__init__(header)
        self.variable_header = variable_header
        self.payload = None

    @classmethod
    def build(cls, packet_id: int):
        v_header = PacketIdVariableHeader(packet_id)
        packet = PubackPacket(variable_header=v_header)
        return packet


class PubcompPacket(MQTTPacket):
    VARIABLE_HEADER = PacketIdVariableHeader
    PAYLOAD = None

    @property
    def packet_id(self):
        return self.variable_header.packet_id

    @packet_id.setter
    def packet_id(self, val: int):
        self.variable_header.packet_id = val

    def __init__(self, fixed: MQTTFixedHeader = None, variable_header: PacketIdVariableHeader = None):
        if fixed is None:
            header = MQTTFixedHeader(PUBCOMP, 0x00)
        else:
            if fixed.packet_type is not PUBCOMP:
                raise HZMQTTException("Invalid fixed packet type %s for PubcompPacket init" % fixed.packet_type)
            header = fixed
        super().__init__(header)
        self.variable_header = variable_header
        self.payload = None

    @classmethod
    def build(cls, packet_id: int):
        v_header = PacketIdVariableHeader(packet_id)
        packet = PubcompPacket(variable_header=v_header)
        return packet


class PublishVariableHeader(MQTTVariableHeader):
    __slots__ = ('topic_name', 'packet_id')

    def __init__(self, topic_name: str, packet_id: int = None):
        super().__init__()
        if '*' in topic_name:
            raise MQTTException("[MQTT-3.3.2-2] Topic name in the PUBLISH Packet MUST NOT contain wildcard characters.")
        self.topic_name = topic_name
        self.packet_id = packet_id

    def __repr__(self):
        return type(self).__name__ + '(topic={0}, packet_id={1})'.format(self.topic_name, self.packet_id)

    def to_bytes(self):
        out = bytearray()
        out.extend(encode_string(self.topic_name))
        if self.packet_id is not None:
            out.extend(int_to_bytes(self.packet_id, 2))
        return out

    @classmethod
    async def from_stream(cls, reader: asyncio.StreamReader, fixed_header: MQTTFixedHeader):
        topic_name = await decode_string(reader)
        has_qos = (fixed_header.flags >> 1) & 0x03
        if has_qos:
            packet_id = await decode_packet_id(reader)
        else:
            packet_id = None
        return cls(topic_name, packet_id)


class PublishPayload(MQTTPayload):
    __slots__ = ('data',)

    def __init__(self, data: bytes = None):
        super().__init__()
        self.data = data

    def to_bytes(self, fixed_header: MQTTFixedHeader, variable_header: MQTTVariableHeader):
        return self.data

    @classmethod
    async def from_stream(cls, reader: asyncio.StreamReader, fixed_header: MQTTFixedHeader, variable_header: MQTTVariableHeader):
        data = bytearray()
        data_length = fixed_header.remaining_length - variable_header.bytes_length
        length_read = 0
        while length_read < data_length:
            buffer = await reader.read(data_length - length_read)
            data.extend(buffer)
            length_read = len(data)
        return cls(data)

    def __repr__(self):
        return type(self).__name__ + '(data={0!r})'.format(repr(self.data))


class PublishPacket(MQTTPacket):
    VARIABLE_HEADER = PublishVariableHeader
    PAYLOAD = PublishPayload

    DUP_FLAG = 0x08
    RETAIN_FLAG = 0x01
    QOS_FLAG = 0x06

    def __init__(self, fixed: MQTTFixedHeader = None, variable_header: PublishVariableHeader = None, payload=None):
        if fixed is None:
            header = MQTTFixedHeader(PUBLISH, 0x00)
        else:
            if fixed.packet_type is not PUBLISH:
                raise HZMQTTException("Invalid fixed packet type %s for PublishPacket init" % fixed.packet_type)
            header = fixed

        super().__init__(header)
        self.variable_header = variable_header
        self.payload = payload

    def set_flags(self, dup_flag=False, qos=0, retain_flag=False):
        self.dup_flag = dup_flag
        self.retain_flag = retain_flag
        self.qos = qos

    def _set_header_flag(self, val, mask):
        if val:
            self.fixed_header.flags |= mask
        else:
            self.fixed_header.flags &= ~mask

    def _get_header_flag(self, mask):
        if self.fixed_header.flags & mask:
            return True
        else:
            return False

    @property
    def dup_flag(self) -> bool:
        return self._get_header_flag(self.DUP_FLAG)

    @dup_flag.setter
    def dup_flag(self, val: bool):
        self._set_header_flag(val, self.DUP_FLAG)

    @property
    def retain_flag(self) -> bool:
        return self._get_header_flag(self.RETAIN_FLAG)

    @retain_flag.setter
    def retain_flag(self, val: bool):
        self._set_header_flag(val, self.RETAIN_FLAG)

    @property
    def qos(self):
        return (self.fixed_header.flags & self.QOS_FLAG) >> 1

    @qos.setter
    def qos(self, val: int):
        self.fixed_header.flags &= 0xf9
        self.fixed_header.flags |= (val << 1)

    @property
    def packet_id(self):
        return self.variable_header.packet_id

    @packet_id.setter
    def packet_id(self, val: int):
        self.variable_header.packet_id = val

    @property
    def data(self):
        return self.payload.data

    @data.setter
    def data(self, data: bytes):
        self.payload.data = data

    @property
    def topic_name(self):
        return self.variable_header.topic_name

    @topic_name.setter
    def topic_name(self, name: str):
        self.variable_header.topic_name = name

    @classmethod
    def build(cls, topic_name: str, message: bytes, packet_id: int, dup_flag, qos, retain):
        v_header = PublishVariableHeader(topic_name, packet_id)
        payload = PublishPayload(message)
        packet = PublishPacket(variable_header=v_header, payload=payload)
        packet.dup_flag = dup_flag
        packet.retain_flag = retain
        packet.qos = qos
        return packet


class PubrecPacket(MQTTPacket):
    VARIABLE_HEADER = PacketIdVariableHeader
    PAYLOAD = None

    @property
    def packet_id(self):
        return self.variable_header.packet_id

    @packet_id.setter
    def packet_id(self, val: int):
        self.variable_header.packet_id = val

    def __init__(self, fixed: MQTTFixedHeader = None, variable_header: PacketIdVariableHeader = None):
        if fixed is None:
            header = MQTTFixedHeader(PUBREC, 0x00)
        else:
            if fixed.packet_type is not PUBREC:
                raise HZMQTTException("Invalid fixed packet type %s for PubrecPacket init" % fixed.packet_type)
            header = fixed
        super().__init__(header)
        self.variable_header = variable_header
        self.payload = None

    @classmethod
    def build(cls, packet_id: int):
        v_header = PacketIdVariableHeader(packet_id)
        packet = PubrecPacket(variable_header=v_header)
        return packet


class PubrelPacket(MQTTPacket):
    VARIABLE_HEADER = PacketIdVariableHeader
    PAYLOAD = None

    @property
    def packet_id(self):
        return self.variable_header.packet_id

    @packet_id.setter
    def packet_id(self, val: int):
        self.variable_header.packet_id = val

    def __init__(self, fixed: MQTTFixedHeader = None, variable_header: PacketIdVariableHeader = None):
        if fixed is None:
            header = MQTTFixedHeader(PUBREL, 0x02)
        else:
            if fixed.packet_type is not PUBREL:
                raise HZMQTTException("Invalid fixed packet type %s for PubrelPacket init" % fixed.packet_type)
            header = fixed
        super().__init__(header)
        self.variable_header = variable_header
        self.payload = None

    @classmethod
    def build(cls, packet_id):
        variable_header = PacketIdVariableHeader(packet_id)
        return PubrelPacket(variable_header=variable_header)


class SubackPayload(MQTTPayload):
    __slots__ = ('return_codes',)

    RETURN_CODE_00 = 0x00
    RETURN_CODE_01 = 0x01
    RETURN_CODE_02 = 0x02
    RETURN_CODE_80 = 0x80

    def __init__(self, return_codes=[]):
        super().__init__()
        self.return_codes = return_codes

    def __repr__(self):
        return type(self).__name__ + '(return_codes={0})'.format(repr(self.return_codes))

    def to_bytes(self, fixed_header: MQTTFixedHeader, variable_header: MQTTVariableHeader):
        out = b''
        for return_code in self.return_codes:
            out += int_to_bytes(return_code, 1)
        return out

    @classmethod
    async def from_stream(cls, reader: ReaderAdapter, fixed_header: MQTTFixedHeader, variable_header: MQTTVariableHeader):
        return_codes = []
        bytes_to_read = fixed_header.remaining_length - variable_header.bytes_length
        for i in range(0, bytes_to_read):
            try:
                return_code_byte = await read_or_raise(reader, 1)
                return_code = bytes_to_int(return_code_byte)
                return_codes.append(return_code)
            except NoDataException:
                break
        return cls(return_codes)


class SubackPacket(MQTTPacket):
    VARIABLE_HEADER = PacketIdVariableHeader
    PAYLOAD = SubackPayload

    def __init__(self, fixed: MQTTFixedHeader = None, variable_header: PacketIdVariableHeader = None, payload=None):
        if fixed is None:
            header = MQTTFixedHeader(SUBACK, 0x00)
        else:
            if fixed.packet_type is not SUBACK:
                raise HZMQTTException("Invalid fixed packet type %s for SubackPacket init" % fixed.packet_type)
            header = fixed

        super().__init__(header)
        self.variable_header = variable_header
        self.payload = payload

    @classmethod
    def build(cls, packet_id, return_codes):
        variable_header = cls.VARIABLE_HEADER(packet_id)
        payload = cls.PAYLOAD(return_codes)
        return cls(variable_header=variable_header, payload=payload)


class SubscribePayload(MQTTPayload):
    __slots__ = ('topics',)

    def __init__(self, topics=[]):
        super().__init__()
        self.topics = topics

    def to_bytes(self, fixed_header: MQTTFixedHeader, variable_header: MQTTVariableHeader):
        out = b''
        for topic in self.topics:
            out += encode_string(topic[0])
            out += int_to_bytes(topic[1], 1)
        return out

    @classmethod
    async def from_stream(cls, reader: asyncio.StreamReader, fixed_header: MQTTFixedHeader, variable_header: MQTTVariableHeader):
        topics = []
        payload_length = fixed_header.remaining_length - variable_header.bytes_length
        read_bytes = 0
        while read_bytes < payload_length:
            try:
                topic = await decode_string(reader)
                qos_byte = await read_or_raise(reader, 1)
                qos = bytes_to_int(qos_byte)
                topics.append((topic, qos))
                read_bytes += 2 + len(topic.encode('utf-8')) + 1
            except NoDataException as exc:
                break
        return cls(topics)

    def __repr__(self):
        return type(self).__name__ + '(topics={0!r})'.format(self.topics)


class SubscribePacket(MQTTPacket):
    VARIABLE_HEADER = PacketIdVariableHeader
    PAYLOAD = SubscribePayload

    def __init__(self, fixed: MQTTFixedHeader = None, variable_header: PacketIdVariableHeader = None, payload=None):
        if fixed is None:
            header = MQTTFixedHeader(SUBSCRIBE, 0x02)  # [MQTT-3.8.1-1]
        else:
            if fixed.packet_type is not SUBSCRIBE:
                raise HZMQTTException("Invalid fixed packet type %s for SubscribePacket init" % fixed.packet_type)
            header = fixed

        super().__init__(header)
        self.variable_header = variable_header
        self.payload = payload

    @classmethod
    def build(cls, topics, packet_id):
        v_header = PacketIdVariableHeader(packet_id)
        payload = SubscribePayload(topics)
        return SubscribePacket(variable_header=v_header, payload=payload)


class UnsubackPacket(MQTTPacket):
    VARIABLE_HEADER = PacketIdVariableHeader
    PAYLOAD = None

    def __init__(self, fixed: MQTTFixedHeader = None, variable_header: PacketIdVariableHeader = None, payload=None):
        if fixed is None:
            header = MQTTFixedHeader(UNSUBACK, 0x00)
        else:
            if fixed.packet_type is not UNSUBACK:
                raise HZMQTTException("Invalid fixed packet type %s for UnsubackPacket init" % fixed.packet_type)
            header = fixed

        super().__init__(header)
        self.variable_header = variable_header
        self.payload = payload

    @classmethod
    def build(cls, packet_id):
        variable_header = PacketIdVariableHeader(packet_id)
        return cls(variable_header=variable_header)


class UnubscribePayload(MQTTPayload):
    __slots__ = ('topics',)

    def __init__(self, topics=[]):
        super().__init__()
        self.topics = topics

    def to_bytes(self, fixed_header: MQTTFixedHeader, variable_header: MQTTVariableHeader):
        out = b''
        for topic in self.topics:
            out += encode_string(topic)
        return out

    @classmethod
    async def from_stream(
            cls, reader: asyncio.StreamReader, fixed_header: MQTTFixedHeader,
            variable_header: MQTTVariableHeader
    ):
        topics = []
        payload_length = fixed_header.remaining_length - variable_header.bytes_length
        read_bytes = 0
        while read_bytes < payload_length:
            try:
                topic = await decode_string(reader)
                topics.append(topic)
                read_bytes += 2 + len(topic.encode('utf-8'))
            except NoDataException:
                break
        return cls(topics)


class UnsubscribePacket(MQTTPacket):
    VARIABLE_HEADER = PacketIdVariableHeader
    PAYLOAD = UnubscribePayload

    def __init__(self, fixed: MQTTFixedHeader = None, variable_header: PacketIdVariableHeader = None, payload=None):
        if fixed is None:
            header = MQTTFixedHeader(UNSUBSCRIBE, 0x02)
        else:
            if fixed.packet_type is not UNSUBSCRIBE:
                raise HZMQTTException("Invalid fixed packet type %s for UnsubscribePacket init" % fixed.packet_type)
            header = fixed

        super().__init__(header)
        self.variable_header = variable_header
        self.payload = payload

    @classmethod
    def build(cls, topics, packet_id):
        v_header = PacketIdVariableHeader(packet_id)
        payload = UnubscribePayload(topics)
        return UnsubscribePacket(variable_header=v_header, payload=payload)



import asyncio
import io
import logging
import math
import sys
import zlib
from asyncio import (
    StreamReader,
    StreamWriter
)
from struct import (
    pack,
    unpack
)

from websockets.exceptions import ConnectionClosed
from websockets.protocol import WebSocketCommonProtocol


class ReaderAdapter:
    async def read(self, n=-1) -> bytes:
        ...

    def feed_eof(self):
        ...


class WriterAdapter:
    def write(self, data):
        ...

    async def drain(self):
        ...

    def get_peer_info(self):
        ...

    async def close(self):
        ...


class WebSocketsReader(ReaderAdapter):
    def __init__(self, protocol: WebSocketCommonProtocol):
        self._protocol = protocol
        self._stream = io.BytesIO(b'')

    async def read(self, n=-1) -> bytes:
        await self._feed_buffer(n)
        data = self._stream.read(n)
        return data

    async def _feed_buffer(self, n=1):
        buffer = bytearray(self._stream.read())
        while len(buffer) < n:
            try:
                message = await self._protocol.recv()
            except ConnectionClosed:
                message = None
            if message is None:
                break
            if not isinstance(message, bytes):
                raise TypeError("message must be bytes")
            buffer.extend(message)
        self._stream = io.BytesIO(buffer)


class WebSocketsWriter(WriterAdapter):
    def __init__(self, protocol: WebSocketCommonProtocol):
        self._protocol = protocol
        self._stream = io.BytesIO(b'')

    def write(self, data):
        self._stream.write(data)

    async def drain(self):
        data = self._stream.getvalue()
        if len(data):
            await self._protocol.send(data)
        self._stream = io.BytesIO(b'')

    def get_peer_info(self):
        return self._protocol.remote_address

    async def close(self):
        await self._protocol.close()


class StreamReaderAdapter(ReaderAdapter):
    def __init__(self, reader: StreamReader):
        self._reader = reader

    async def read(self, n=-1) -> bytes:
        if n == -1:
            data = await self._reader.read(n)
        else:
            data = await self._reader.readexactly(n)
        return data

    def feed_eof(self):
        return self._reader.feed_eof()


class StreamWriterAdapter(WriterAdapter):
    def __init__(self, writer: StreamWriter):
        self.logger = logging.getLogger(__name__)
        self._writer = writer
        self.is_closed = False

    def write(self, data):
        if not self.is_closed:
            self._writer.write(data)

    async def drain(self):
        if not self.is_closed:
            await self._writer.drain()

    def get_peer_info(self):
        extra_info = self._writer.get_extra_info('peername')
        return extra_info[0], extra_info[1]

    async def close(self):
        if not self.is_closed:
            self.is_closed = True
            await self._writer.drain()
            if self._writer.can_write_eof():
                self._writer.write_eof()
            self._writer.close()
            try:
                await self._writer.wait_closed()
            except AttributeError:
                pass


class BufferReader(ReaderAdapter):
    def __init__(self, buffer: bytes):
        self._stream = io.BytesIO(buffer)

    async def read(self, n=-1) -> bytes:
        return self._stream.read(n)


class BufferWriter(WriterAdapter):
    def __init__(self, buffer=b''):
        self._stream = io.BytesIO(buffer)

    def write(self, data):
        self._stream.write(data)

    async def drain(self):
        pass

    def get_buffer(self):
        return self._stream.getvalue()

    def get_peer_info(self):
        return "BufferWriter", 0

    async def close(self):
        self._stream.close()


ECB = 0
CBC = 1

PAD_NORMAL = 1
PAD_PKCS5 = 2


class BaseDES(object):
    def __init__(self, mode=ECB, iv=None, pad=None, padmode=PAD_NORMAL):
        if iv:
            iv = self._guard_against_unicode(iv)
        if pad:
            pad = self._guard_against_unicode(pad)
        self.block_size = 8

        if pad and padmode == PAD_PKCS5:
            raise ValueError("Cannot use a pad character with PAD_PKCS5")
        if iv and len(iv) != self.block_size:
            raise ValueError("Invalid Initial Value (IV), must be a multiple of " + str(self.block_size) + " bytes")

        self._mode = mode
        self._iv = iv
        self._padding = pad
        self._padmode = padmode

        self.__key = b''

    def get_key(self):
        return self.__key

    def set_key(self, key):
        key = self._guard_against_unicode(key)
        self.__key = key

    def get_mode(self):
        return self._mode

    def set_mode(self, mode):
        self._mode = mode

    def get_padding(self):
        return self._padding

    def set_padding(self, pad):
        if pad is not None:
            pad = self._guard_against_unicode(pad)
        self._padding = pad

    def get_pad_mode(self):
        return self._padmode

    def set_pad_mode(self, mode):
        self._padmode = mode

    def get_iv(self):
        return self._iv

    def set_iv(self, iv):
        if not iv or len(iv) != self.block_size:
            raise ValueError("Invalid Initial Value (IV), must be a multiple of " + str(self.block_size) + " bytes")
        iv = self._guard_against_unicode(iv)
        self._iv = iv

    def _pad_data(self, data, pad, padmode):
        if padmode is None:
            padmode = self.get_pad_mode()
        if pad and padmode == PAD_PKCS5:
            raise ValueError("Cannot use a pad character with PAD_PKCS5")

        if padmode == PAD_NORMAL:
            if len(data) % self.block_size == 0:
                return data

            if not pad:
                pad = self.get_padding()
            if not pad:
                raise ValueError("Data must be a multiple of " + str(self.block_size) + " bytes in length. Use padmode=PAD_PKCS5 or set the pad character.")
            data += (self.block_size - (len(data) % self.block_size)) * pad

        elif padmode == PAD_PKCS5:
            pad_len = 8 - (len(data) % self.block_size)
            data += bytes([pad_len] * pad_len)

        return data

    def _unpad_data(self, data, pad, padmode):
        if not data:
            return data
        if pad and padmode == PAD_PKCS5:
            raise ValueError("Cannot use a pad character with PAD_PKCS5")
        if padmode is None:
            padmode = self.get_pad_mode()

        if padmode == PAD_NORMAL:
            if not pad:
                pad = self.get_padding()
            if pad:
                data = data[:-self.block_size] + data[-self.block_size:].rstrip(pad)

        elif padmode == PAD_PKCS5:
            pad_len = data[-1]
            data = data[:-pad_len]

        return data

    @staticmethod
    def _guard_against_unicode(data):
        if isinstance(data, str):
            try:
                return data.encode('ascii')
            except UnicodeEncodeError:
                pass
            raise ValueError("pyDes can only work with encoded strings, not Unicode.")
        return data


class DES(BaseDES):
    __pc1 = [56, 48, 40, 32, 24, 16, 8,
             0, 57, 49, 41, 33, 25, 17,
             9, 1, 58, 50, 42, 34, 26,
             18, 10, 2, 59, 51, 43, 35,
             62, 54, 46, 38, 30, 22, 14,
             6, 61, 53, 45, 37, 29, 21,
             13, 5, 60, 52, 44, 36, 28,
             20, 12, 4, 27, 19, 11, 3
             ]

    __left_rotations = [
        1, 1, 2, 2, 2, 2, 2, 2, 1, 2, 2, 2, 2, 2, 2, 1
    ]

    __pc2 = [
        13, 16, 10, 23, 0, 4,
        2, 27, 14, 5, 20, 9,
        22, 18, 11, 3, 25, 7,
        15, 6, 26, 19, 12, 1,
        40, 51, 30, 36, 46, 54,
        29, 39, 50, 44, 32, 47,
        43, 48, 38, 55, 33, 52,
        45, 41, 49, 35, 28, 31
    ]

    __ip = [57, 49, 41, 33, 25, 17, 9, 1,
            59, 51, 43, 35, 27, 19, 11, 3,
            61, 53, 45, 37, 29, 21, 13, 5,
            63, 55, 47, 39, 31, 23, 15, 7,
            56, 48, 40, 32, 24, 16, 8, 0,
            58, 50, 42, 34, 26, 18, 10, 2,
            60, 52, 44, 36, 28, 20, 12, 4,
            62, 54, 46, 38, 30, 22, 14, 6
            ]

    __expansion_table = [
        31, 0, 1, 2, 3, 4,
        3, 4, 5, 6, 7, 8,
        7, 8, 9, 10, 11, 12,
        11, 12, 13, 14, 15, 16,
        15, 16, 17, 18, 19, 20,
        19, 20, 21, 22, 23, 24,
        23, 24, 25, 26, 27, 28,
        27, 28, 29, 30, 31, 0
    ]

    __sbox = [
        [14, 4, 13, 1, 2, 15, 11, 8, 3, 10, 6, 12, 5, 9, 0, 7,
         0, 15, 7, 4, 14, 2, 13, 1, 10, 6, 12, 11, 9, 5, 3, 8,
         4, 1, 14, 8, 13, 6, 2, 11, 15, 12, 9, 7, 3, 10, 5, 0,
         15, 12, 8, 2, 4, 9, 1, 7, 5, 11, 3, 14, 10, 0, 6, 13],

        [15, 1, 8, 14, 6, 11, 3, 4, 9, 7, 2, 13, 12, 0, 5, 10,
         3, 13, 4, 7, 15, 2, 8, 14, 12, 0, 1, 10, 6, 9, 11, 5,
         0, 14, 7, 11, 10, 4, 13, 1, 5, 8, 12, 6, 9, 3, 2, 15,
         13, 8, 10, 1, 3, 15, 4, 2, 11, 6, 7, 12, 0, 5, 14, 9],

        [10, 0, 9, 14, 6, 3, 15, 5, 1, 13, 12, 7, 11, 4, 2, 8,
         13, 7, 0, 9, 3, 4, 6, 10, 2, 8, 5, 14, 12, 11, 15, 1,
         13, 6, 4, 9, 8, 15, 3, 0, 11, 1, 2, 12, 5, 10, 14, 7,
         1, 10, 13, 0, 6, 9, 8, 7, 4, 15, 14, 3, 11, 5, 2, 12],

        [7, 13, 14, 3, 0, 6, 9, 10, 1, 2, 8, 5, 11, 12, 4, 15,
         13, 8, 11, 5, 6, 15, 0, 3, 4, 7, 2, 12, 1, 10, 14, 9,
         10, 6, 9, 0, 12, 11, 7, 13, 15, 1, 3, 14, 5, 2, 8, 4,
         3, 15, 0, 6, 10, 1, 13, 8, 9, 4, 5, 11, 12, 7, 2, 14],

        [2, 12, 4, 1, 7, 10, 11, 6, 8, 5, 3, 15, 13, 0, 14, 9,
         14, 11, 2, 12, 4, 7, 13, 1, 5, 0, 15, 10, 3, 9, 8, 6,
         4, 2, 1, 11, 10, 13, 7, 8, 15, 9, 12, 5, 6, 3, 0, 14,
         11, 8, 12, 7, 1, 14, 2, 13, 6, 15, 0, 9, 10, 4, 5, 3],

        [12, 1, 10, 15, 9, 2, 6, 8, 0, 13, 3, 4, 14, 7, 5, 11,
         10, 15, 4, 2, 7, 12, 9, 5, 6, 1, 13, 14, 0, 11, 3, 8,
         9, 14, 15, 5, 2, 8, 12, 3, 7, 0, 4, 10, 1, 13, 11, 6,
         4, 3, 2, 12, 9, 5, 15, 10, 11, 14, 1, 7, 6, 0, 8, 13],

        [4, 11, 2, 14, 15, 0, 8, 13, 3, 12, 9, 7, 5, 10, 6, 1,
         13, 0, 11, 7, 4, 9, 1, 10, 14, 3, 5, 12, 2, 15, 8, 6,
         1, 4, 11, 13, 12, 3, 7, 14, 10, 15, 6, 8, 0, 5, 9, 2,
         6, 11, 13, 8, 1, 4, 10, 7, 9, 5, 0, 15, 14, 2, 3, 12],

        [13, 2, 8, 4, 6, 15, 11, 1, 10, 9, 3, 14, 5, 0, 12, 7,
         1, 15, 13, 8, 10, 3, 7, 4, 12, 5, 6, 11, 0, 14, 9, 2,
         7, 11, 4, 1, 9, 12, 14, 2, 0, 6, 10, 13, 15, 3, 5, 8,
         2, 1, 14, 7, 4, 10, 8, 13, 15, 12, 9, 0, 3, 5, 6, 11],
    ]

    __p = [
        15, 6, 19, 20, 28, 11,
        27, 16, 0, 14, 22, 25,
        4, 17, 30, 9, 1, 7,
        23, 13, 31, 26, 2, 8,
        18, 12, 29, 5, 21, 10,
        3, 24
    ]

    __fp = [
        39, 7, 47, 15, 55, 23, 63, 31,
        38, 6, 46, 14, 54, 22, 62, 30,
        37, 5, 45, 13, 53, 21, 61, 29,
        36, 4, 44, 12, 52, 20, 60, 28,
        35, 3, 43, 11, 51, 19, 59, 27,
        34, 2, 42, 10, 50, 18, 58, 26,
        33, 1, 41, 9, 49, 17, 57, 25,
        32, 0, 40, 8, 48, 16, 56, 24
    ]

    ENCRYPT = 0x00
    DECRYPT = 0x01

    def __init__(self, key, mode=ECB, iv=None, pad=None, padmode=PAD_NORMAL):
        if len(key) != 8:
            raise ValueError("Invalid DES key size. Key must be exactly 8 bytes long.")
        BaseDES.__init__(self, mode, iv, pad, padmode)
        self.key_size = 8

        self.L = []
        self.R = []
        self.Kn = [[0] * 48] * 16
        self.final = []

        self.set_key(key)

    def set_key(self, key):
        BaseDES.set_key(self, key)
        self.__create_sub_keys()

    @staticmethod
    def __string_to_bit_list(data):
        if isinstance(data[0], str):
            data = [ord(c) for c in data]
        length = len(data) * 8
        result = [0] * length
        pos = 0
        for ch in data:
            i = 7
            while i >= 0:
                if ch & (1 << i) != 0:
                    result[pos] = 1
                else:
                    result[pos] = 0
                pos += 1
                i -= 1

        return result

    @staticmethod
    def __bit_list_to_string(data):
        result = []
        pos = 0
        c = 0
        while pos < len(data):
            c += data[pos] << (7 - (pos % 8))
            if (pos % 8) == 7:
                result.append(c)
                c = 0
            pos += 1

        return bytes(result)

    @staticmethod
    def __permutate(table, block):
        return list(map(lambda x: block[x], table))

    def __create_sub_keys(self):
        key = self.__permutate(DES.__pc1, self.__string_to_bit_list(self.get_key()))
        i = 0
        self.L = key[:28]
        self.R = key[28:]
        while i < 16:
            j = 0
            while j < DES.__left_rotations[i]:
                self.L.append(self.L[0])
                del self.L[0]

                self.R.append(self.R[0])
                del self.R[0]

                j += 1

            self.Kn[i] = self.__permutate(DES.__pc2, self.L + self.R)

            i += 1

    def __des_crypt(self, block, crypt_type):
        block = self.__permutate(DES.__ip, block)
        self.L = block[:32]
        self.R = block[32:]

        if crypt_type == DES.ENCRYPT:
            iteration = 0
            iteration_adjustment = 1
        else:
            iteration = 15
            iteration_adjustment = -1

        i = 0
        while i < 16:
            temp_r = self.R[:]

            self.R = self.__permutate(DES.__expansion_table, self.R)

            self.R = list(map(lambda x, y: x ^ y, self.R, self.Kn[iteration]))
            b = [self.R[:6], self.R[6:12], self.R[12:18], self.R[18:24], self.R[24:30], self.R[30:36], self.R[36:42], self.R[42:]]

            j = 0
            bn = [0] * 32
            pos = 0
            while j < 8:
                m = (b[j][0] << 1) + b[j][5]
                n = (b[j][1] << 3) + (b[j][2] << 2) + (b[j][3] << 1) + b[j][4]

                v = DES.__sbox[j][(m << 4) + n]

                bn[pos] = (v & 8) >> 3
                bn[pos + 1] = (v & 4) >> 2
                bn[pos + 2] = (v & 2) >> 1
                bn[pos + 3] = v & 1

                pos += 4
                j += 1

            self.R = self.__permutate(DES.__p, bn)

            self.R = list(map(lambda x, y: x ^ y, self.R, self.L))
            self.L = temp_r

            i += 1
            iteration += iteration_adjustment

        self.final = self.__permutate(DES.__fp, self.R + self.L)
        return self.final

    def crypt(self, data, crypt_type):
        if not data:
            return ''
        if len(data) % self.block_size != 0:
            if crypt_type == DES.DECRYPT:
                raise ValueError("Invalid data length, data must be a multiple of " + str(self.block_size) + " bytes\n.")
            if not self.get_padding():
                raise ValueError("Invalid data length, data must be a multiple of " + str(self.block_size) + " bytes\n. Try setting the optional padding character")
            else:
                data += (self.block_size - (len(data) % self.block_size)) * self.get_padding()

        if self.get_mode() == CBC:
            if self.get_iv():
                iv = self.__string_to_bit_list(self.get_iv())
            else:
                raise ValueError("For CBC mode, you must supply the Initial Value (IV) for ciphering")

        i = 0
        result = []
        while i < len(data):
            block = self.__string_to_bit_list(data[i:i + 8])

            if self.get_mode() == CBC:
                if crypt_type == DES.ENCRYPT:
                    block = list(map(lambda x, y: x ^ y, block, iv))

                processed_block = self.__des_crypt(block, crypt_type)

                if crypt_type == DES.DECRYPT:
                    processed_block = list(map(lambda x, y: x ^ y, processed_block, iv))
                    iv = block
                else:
                    iv = processed_block
            else:
                processed_block = self.__des_crypt(block, crypt_type)

            result.append(self.__bit_list_to_string(processed_block))
            i += 8

        return bytes.fromhex('').join(result)

    def encrypt(self, data, pad=None, padmode=None):
        data = self._guard_against_unicode(data)
        if pad is not None:
            pad = self._guard_against_unicode(pad)
        data = self._pad_data(data, pad, padmode)
        return self.crypt(data, DES.ENCRYPT)

    def decrypt(self, data, pad=None, padmode=None):
        data = self._guard_against_unicode(data)
        if pad is not None:
            pad = self._guard_against_unicode(pad)
        data = self.crypt(data, DES.DECRYPT)
        return self._unpad_data(data, pad, padmode)


class TripleDES(BaseDES):
    def __init__(self, key, mode=ECB, iv=None, pad=None, padmode=PAD_NORMAL):
        BaseDES.__init__(self, mode, iv, pad, padmode)
        self.set_key(key)

        self.key_size = 0
        self.__key1 = None
        self.__key2 = None
        self.__key3 = None

    def set_key(self, key):
        self.key_size = 24
        if len(key) != self.key_size:
            if len(key) == 16:
                self.key_size = 16
            else:
                raise ValueError("Invalid triple DES key size. Key must be either 16 or 24 bytes long")
        if self.get_mode() == CBC:
            if not self.get_iv():
                self._iv = key[:self.block_size]
            if len(self.get_iv()) != self.block_size:
                raise ValueError("Invalid IV, must be 8 bytes in length")
        self.__key1 = DES(key[:8], self._mode, self._iv, self._padding, self._padmode)
        self.__key2 = DES(key[8:16], self._mode, self._iv, self._padding, self._padmode)
        if self.key_size == 16:
            self.__key3 = self.__key1
        else:
            self.__key3 = DES(key[16:], self._mode, self._iv, self._padding, self._padmode)
        BaseDES.set_key(self, key)

    def set_mode(self, mode):
        BaseDES.set_mode(self, mode)
        for key in (self.__key1, self.__key2, self.__key3):
            key.set_mode(mode)

    def set_padding(self, pad):
        BaseDES.set_padding(self, pad)
        for key in (self.__key1, self.__key2, self.__key3):
            key.set_padding(pad)

    def set_pad_mode(self, mode):
        BaseDES.set_pad_mode(self, mode)
        for key in (self.__key1, self.__key2, self.__key3):
            key.set_pad_mode(mode)

    def set_iv(self, iv):
        BaseDES.set_iv(self, iv)
        for key in (self.__key1, self.__key2, self.__key3):
            key.set_iv(iv)

    def encrypt(self, data, pad=None, padmode=None):
        encrypt = DES.ENCRYPT
        decrypt = DES.DECRYPT
        data = self._guard_against_unicode(data)
        if pad is not None:
            pad = self._guard_against_unicode(pad)

        data = self._pad_data(data, pad, padmode)
        if self.get_mode() == CBC:
            self.__key1.set_iv(self.get_iv())
            self.__key2.set_iv(self.get_iv())
            self.__key3.set_iv(self.get_iv())
            i = 0
            result = []
            while i < len(data):
                block = self.__key1.crypt(data[i:i + 8], encrypt)
                block = self.__key2.crypt(block, decrypt)
                block = self.__key3.crypt(block, encrypt)
                self.__key1.set_iv(block)
                self.__key2.set_iv(block)
                self.__key3.set_iv(block)
                result.append(block)
                i += 8
            return bytes.fromhex('').join(result)
        else:
            data = self.__key1.crypt(data, encrypt)
            data = self.__key2.crypt(data, decrypt)
            return self.__key3.crypt(data, encrypt)

    def decrypt(self, data, pad=None, padmode=None):
        encrypt = DES.ENCRYPT
        decrypt = DES.DECRYPT
        data = self._guard_against_unicode(data)
        if pad is not None:
            pad = self._guard_against_unicode(pad)
        if self.get_mode() == CBC:
            self.__key1.set_iv(self.get_iv())
            self.__key2.set_iv(self.get_iv())
            self.__key3.set_iv(self.get_iv())
            i = 0
            result = []
            while i < len(data):
                iv = data[i:i + 8]
                block = self.__key3.crypt(iv, decrypt)
                block = self.__key2.crypt(block, encrypt)
                block = self.__key1.crypt(block, decrypt)
                self.__key1.set_iv(iv)
                self.__key2.set_iv(iv)
                self.__key3.set_iv(iv)
                result.append(block)
                i += 8
            data = bytes.fromhex('').join(result)
        else:
            data = self.__key3.crypt(data, decrypt)
            data = self.__key2.crypt(data, encrypt)
            data = self.__key1.crypt(data, decrypt)
        return self._unpad_data(data, pad, padmode)


if sys.version_info[0] >= 3:
    original_ord = ord


    def ord(x):
        if isinstance(x, int):
            return x
        elif isinstance(x, bytes) or isinstance(x, str):
            return original_ord(x)
        else:
            raise TypeError(f"our customized ord takes an int, a byte, or a str. Got {type(x)} : {x}")

RAW_ENCODING = 0
COPY_RECTANGLE_ENCODING = 1
RRE_ENCODING = 2
CORRE_ENCODING = 4
HEXTILE_ENCODING = 5
ZLIB_ENCODING = 6
TIGHT_ENCODING = 7
ZLIBHEX_ENCODING = 8
ZRLE_ENCODING = 16
PSEUDO_CURSOR_ENCODING = -239
PSEUDO_DESKTOP_SIZE_ENCODING = -223

KEY_BackSpace = 0xff08
KEY_Tab = 0xff09
KEY_Return = 0xff0d
KEY_Escape = 0xff1b
KEY_Insert = 0xff63
KEY_Delete = 0xffff
KEY_Home = 0xff50
KEY_End = 0xff57
KEY_PageUp = 0xff55
KEY_PageDown = 0xff56
KEY_Left = 0xff51
KEY_Up = 0xff52
KEY_Right = 0xff53
KEY_Down = 0xff54
KEY_F1 = 0xffbe
KEY_F2 = 0xffbf
KEY_F3 = 0xffc0
KEY_F4 = 0xffc1
KEY_F5 = 0xffc2
KEY_F6 = 0xffc3
KEY_F7 = 0xffc4
KEY_F8 = 0xffc5
KEY_F9 = 0xffc6
KEY_F10 = 0xffc7
KEY_F11 = 0xffc8
KEY_F12 = 0xffc9
KEY_F13 = 0xFFCA
KEY_F14 = 0xFFCB
KEY_F15 = 0xFFCC
KEY_F16 = 0xFFCD
KEY_F17 = 0xFFCE
KEY_F18 = 0xFFCF
KEY_F19 = 0xFFD0
KEY_F20 = 0xFFD1
KEY_ShiftLeft = 0xffe1
KEY_ShiftRight = 0xffe2
KEY_ControlLeft = 0xffe3
KEY_ControlRight = 0xffe4
KEY_MetaLeft = 0xffe7
KEY_MetaRight = 0xffe8
KEY_AltLeft = 0xffe9
KEY_AltRight = 0xffea

KEY_Scroll_Lock = 0xFF14
KEY_Sys_Req = 0xFF15
KEY_Num_Lock = 0xFF7F
KEY_Caps_Lock = 0xFFE5
KEY_Pause = 0xFF13
KEY_Super_L = 0xFFEB
KEY_Super_R = 0xFFEC
KEY_Hyper_L = 0xFFED
KEY_Hyper_R = 0xFFEE

KEY_KP_0 = 0xFFB0
KEY_KP_1 = 0xFFB1
KEY_KP_2 = 0xFFB2
KEY_KP_3 = 0xFFB3
KEY_KP_4 = 0xFFB4
KEY_KP_5 = 0xFFB5
KEY_KP_6 = 0xFFB6
KEY_KP_7 = 0xFFB7
KEY_KP_8 = 0xFFB8
KEY_KP_9 = 0xFFB9
KEY_KP_Enter = 0xFF8D

KEY_ForwardSlash = 0x002F
KEY_BackSlash = 0x005C
KEY_SpaceBar = 0x0020


def _zrle_next_bit(it, pixels_in_tile):
    num_pixels = 0
    while True:
        b = ord(next(it))

        for n in range(8):
            value = b >> (7 - n)
            yield value & 1

            num_pixels += 1
            if num_pixels == pixels_in_tile:
                return


def _zrle_next_dibit(it, pixels_in_tile):
    num_pixels = 0
    while True:
        b = ord(next(it))

        for n in range(0, 8, 2):
            value = b >> (6 - n)
            yield value & 3

            num_pixels += 1
            if num_pixels == pixels_in_tile:
                return


def _zrle_next_nibble(it, pixels_in_tile):
    num_pixels = 0
    while True:
        b = ord(next(it))

        for n in range(0, 8, 4):
            value = b >> (4 - n)
            yield value & 15

            num_pixels += 1
            if num_pixels == pixels_in_tile:
                return


class RFBClient:
    def __init__(self, host, port, loop):
        self.host = host
        self.port = port
        self.loop = loop

        self.reader = None
        self.writer = None

        self.version = None
        self.server_version = None
        self.password = None
        self.shared = 0

        self._zlib_stream = zlib.decompressobj(0)

    async def connect(self):
        conn_reader, conn_writer = await asyncio.open_connection(self.host, self.port, loop=self.loop)
        self.reader = StreamReaderAdapter(conn_reader)
        self.writer = StreamWriterAdapter(conn_writer)

        buffer = b''
        while True:
            byte = await self.reader.read(1)
            buffer += byte
            if byte == b'\n':
                break

        version = 3.3
        if buffer[:3] == b'RFB':
            version_server = float(buffer[3:-1].replace(b'0', b''))
            self.server_version = version_server
            supported_versions = (3.3, 3.7, 3.8)
            if version_server in supported_versions:
                version = version_server
            else:
                print("Protocol version %.3f not supported" % version_server)
                version = max(filter(lambda x: x <= version_server, supported_versions))
        print("Using protocol version %.3f" % version)
        parts = str(version).split('.')
        self.writer.write(bytes(b"RFB %03d.%03d\n" % (int(parts[0]), int(parts[1]))))

        self.version = version
        if version < 3.7:
            await self._handle_auth()
        else:
            await self._handle_number_security_types()

    async def _handle_auth(self):
        buffer = await self.reader.read(4)

        (auth,) = unpack("!I", buffer)
        if auth == 0:
            await self._handle_conn_failed()
        elif auth == 1:
            await self._do_client_initialization()
            return
        elif auth == 2:
            await self._handle_vnc_auth()
        else:
            print("unknown auth response (%d)" % auth)

    async def _handle_number_security_types(self):
        buffer = await self.reader.read(1)
        num_types, = unpack("!B", buffer)
        if num_types:
            num_types = await self.reader.read(num_types)
            types = unpack("!%dB" % len(num_types), num_types)
            supported_types = (1, 2)
            valid_types = [sec_type for sec_type in types if sec_type in supported_types]
            if valid_types:
                sec_type = max(valid_types)
                self.writer.write(pack("!B", sec_type))
                if sec_type == 1:
                    if self.version < 3.8:
                        await self._do_client_initialization()
                    else:
                        await self._handle_vnc_auth_result()
                else:
                    await self._handle_vnc_auth()
            else:
                print("unknown security types: %s" % repr(types))
        else:
            await self._handle_conn_failed()

    async def _handle_vnc_auth(self):
        buffer = await self.reader.read(16)
        self._challenge = buffer
        await self.vnc_request_password()
        await self._handle_vnc_auth_result()

    async def _handle_vnc_auth_result(self):
        buffer = await self.reader.read(4)
        (result,) = unpack("!I", buffer)
        if result == 0:
            await self._do_client_initialization()
            return
        elif result == 1:
            if self.version < 3.8:
                await self.vnc_auth_failed("authentication failed")
                await self.writer.close()
            else:
                await self._handle_auth_failed()
        elif result == 2:
            if self.version < 3.8:
                await self.vnc_auth_failed("too many tries to log in")
                await self.writer.close()
            else:
                await self._handle_auth_failed()
        else:
            print("unknown auth response (%d)" % result)

    async def _do_client_initialization(self):
        self.writer.write(pack("!B", self.shared))

        buffer = await self.reader.read(24)
        self.width, self.height, pixformat, namelen = unpack("!HH16sI", buffer)
        self.bpp, self.depth, self.bigendian, self.truecolor, self.redmax, self.greenmax, self.bluemax, self.redshift, self.greenshift, self.blueshift = unpack(
            "!BBBBHHHBBBxxx",
            pixformat
        )
        self.bypp = self.bpp // 8

        buffer = await self.reader.read(namelen)
        self.name = buffer

        await self.vnc_connection_made()
        # await self._handle_connection()

        asyncio.ensure_future(self._handle_connection(), loop=self.loop)

    async def _handle_auth_failed(self):
        buffer = await self.reader.read(4)
        waitfor, = unpack("!I", buffer)
        await self._handle_auth_failed_message(waitfor)

    async def _handle_auth_failed_message(self, waitfor):
        buffer = await self.reader.read(waitfor)
        await self.vnc_auth_failed(buffer)
        await self.writer.close()

    async def _handle_connection(self):
        while True:
            await asyncio.sleep(0.01, loop=self.loop)

            buffer = await self.reader.read(1)

            msgid, = unpack("!B", buffer)
            if msgid == 0:
                await self._handle_framebuffer_update()
            elif msgid == 2:
                await self.bell()
            elif msgid == 3:
                await self._handle_server_cut_text()
            else:
                print("unknown message received (id %d)" % msgid)

    async def _handle_conn_failed(self):
        buffer = await self.reader.read(4)
        waitfor, = unpack("!I", buffer)
        await self._handle_conn_message(waitfor)

    async def _handle_conn_message(self, waitfor):
        buffer = await self.reader.read(waitfor)
        print("Connection refused: %r" % buffer)

    async def _do_connection(self):
        if self.rectangles:
            await self._handle_rectangle()
        else:
            await self.commit_update(self.rectanglePos)

    async def _handle_rectangle(self):
        buffer = await self.reader.read(12)

        (x, y, width, height, encoding) = unpack("!HHHHi", buffer)
        if self.rectangles:
            self.rectangles -= 1
            self.rectanglePos.append((x, y, width, height))
            if encoding == COPY_RECTANGLE_ENCODING:
                await self._handle_decode_copyrect(x, y, width, height)
            elif encoding == RAW_ENCODING:
                await self._handle_decode_raw(x, y, width, height)
            elif encoding == HEXTILE_ENCODING:
                await self._do_next_hextile_subrect(None, None, x, y, width, height, None, None)
            elif encoding == CORRE_ENCODING:
                await self._handle_decode_corre(x, y, width, height)
            elif encoding == RRE_ENCODING:
                await self._handle_decode_rre(x, y, width, height)
            elif encoding == ZRLE_ENCODING:
                await self._handle_decode_zrle(x, y, width, height)
            elif encoding == PSEUDO_CURSOR_ENCODING:
                length = width * height * self.bypp
                length += int(math.floor((width + 7.0) / 8)) * height
                await self._handle_decode_psuedo_cursor(length, x, y, width, height)
            elif encoding == PSEUDO_DESKTOP_SIZE_ENCODING:
                await self._handle_decode_desktop_size(width, height)
            else:
                print("unknown encoding received (encoding %d)" % encoding)
                await self._do_connection()
        else:
            await self._do_connection()

    async def _handle_decode_copyrect(self, x, y, width, height):
        buffer = await self.reader.read(4)

        (srcx, srcy) = unpack("!HH", buffer)
        await self.copy_rectangle(srcx, srcy, x, y, width, height)
        await self._do_connection()

    async def _handle_decode_raw(self, x, y, width, height):
        buffer = await self.reader.read(width * height * self.bypp)

        await self.update_rectangle(x, y, width, height, buffer)
        await self._do_connection()

    async def _do_next_hextile_subrect(self, bg, color, x, y, width, height, tx, ty):
        if tx is not None:
            tx += 16
            if tx >= x + width:
                tx = x
                ty += 16
        else:
            tx = x
            ty = y

        if ty >= y + height:
            await self._do_connection()
        else:
            await self._handle_decode_hextile(bg, color, x, y, width, height, tx, ty)

    async def _handle_decode_hextile(self, bg, color, x, y, width, height, tx, ty):
        buffer = await self.reader.read(1)
        (subencoding,) = unpack("!B", buffer)

        tw = th = 16
        if x + width - tx < 16:
            tw = x + width - tx
        if y + height - ty < 16:
            th = y + height - ty

        if subencoding & 1:
            await self._handle_decode_hextile_raw(bg, color, x, y, width, height, tx, ty, tw, th)
        else:
            numbytes = 0
            if subencoding & 2:
                numbytes += self.bypp
            if subencoding & 4:
                numbytes += self.bypp
            if subencoding & 8:
                numbytes += 1
            if numbytes:
                await self._handle_decode_hextile_subrect(subencoding, bg, color, x, y, width, height, tx, ty, tw, th, numbytes)
            else:
                await self.fill_rectangle(tx, ty, tw, th, bg)
                await self._do_next_hextile_subrect(bg, color, x, y, width, height, tx, ty)

    async def _handle_decode_hextile_raw(self, bg, color, x, y, width, height, tx, ty, tw, th):
        buffer = self.reader.read(tw * th * self.bypp)
        await self.update_rectangle(tx, ty, tw, th, buffer)
        await self._do_next_hextile_subrect(bg, color, x, y, width, height, tx, ty)

    async def _handle_decode_hextile_subrect(self, subencoding, bg, color, x, y, width, height, tx, ty, tw, th, numbytes):
        buffer = self.reader.read(numbytes)

        subrects = 0
        pos = 0
        if subencoding & 2:
            bg = buffer[:self.bypp]
            pos += self.bypp
        await self.fill_rectangle(tx, ty, tw, th, bg)
        if subencoding & 4:
            color = buffer[pos:pos + self.bypp]
            pos += self.bypp
        if subencoding & 8:
            subrects = ord(buffer[pos])

        if subrects:
            if subencoding & 16:
                await self._handle_decode_hextile_subrects_coloured(bg, color, subrects, x, y, width, height, tx, ty, tw, th)
            else:
                await self._handle_decode_hextile_subrects_fg(bg, color, subrects, x, y, width, height, tx, ty, tw, th)
        else:
            await self._do_next_hextile_subrect(bg, color, x, y, width, height, tx, ty)

    async def _handle_decode_hextile_subrects_coloured(self, bg, color, subrects, x, y, width, height, tx, ty, tw, th):
        buffer = self.reader.read((self.bypp + 2) * subrects)
        sz = self.bypp + 2
        pos = 0
        end = len(buffer)
        while pos < end:
            pos2 = pos + self.bypp
            color = buffer[pos:pos2]
            xy = ord(buffer[pos2])
            wh = ord(buffer[pos2 + 1])
            sx = xy >> 4
            sy = xy & 0xf
            sw = (wh >> 4) + 1
            sh = (wh & 0xf) + 1
            await self.fill_rectangle(tx + sx, ty + sy, sw, sh, color)
            pos += sz
        await self._do_next_hextile_subrect(bg, color, x, y, width, height, tx, ty)

    async def _handle_decode_hextile_subrects_fg(self, bg, color, subrects, x, y, width, height, tx, ty, tw, th):
        buffer = self.reader.read(2 * subrects)
        pos = 0
        end = len(buffer)
        while pos < end:
            xy = ord(buffer[pos])
            wh = ord(buffer[pos + 1])
            sx = xy >> 4
            sy = xy & 0xf
            sw = (wh >> 4) + 1
            sh = (wh & 0xf) + 1
            await self.fill_rectangle(tx + sx, ty + sy, sw, sh, color)
            pos += 2
        await self._do_next_hextile_subrect(bg, color, x, y, width, height, tx, ty)

    async def _handle_decode_corre(self, x, y, width, height):
        buffer = await self.reader.read(4 + self.bypp)

        (subrects,) = unpack("!I", buffer[:4])
        color = buffer[4:]
        await self.fill_rectangle(x, y, width, height, color)
        if subrects:
            await self._handle_decode_corre_rectangles(x, y, subrects)
        else:
            await self._do_connection()

    async def _handle_decode_corre_rectangles(self, topx, topy, subrects):
        buffer = await self.reader.read((8 + self.bypp) * subrects)

        pos = 0
        end = len(buffer)
        sz = self.bypp + 4
        fmt = "!%dsBBBB" % self.bypp
        while pos < sz:
            (color, x, y, width, height) = unpack(fmt, buffer[pos:pos + sz])
            await self.fill_rectangle(topx + x, topy + y, width, height, color)
            pos += sz
        await self._do_connection()

    async def _handle_decode_rre(self, x, y, width, height):
        buffer = await self.reader.read(4 + self.bypp)
        (subrects,) = unpack("!I", buffer[:4])
        color = buffer[4:]
        await self.fill_rectangle(x, y, width, height, color)
        if subrects:
            await self._handle_rre_sub_rectangles(x, y, subrects)
        else:
            await self._do_connection()

    async def _handle_rre_sub_rectangles(self, topx, topy, subrects):
        buffer = await self.reader.read((8 + self.bypp) * subrects)

        pos = 0
        end = len(buffer)
        sz = self.bypp + 8
        fmt = "!%dsHHHH" % self.bypp
        while pos < end:
            (color, x, y, width, height) = unpack(fmt, buffer[pos:pos + sz])
            await self.fill_rectangle(topx + x, topy + y, width, height, color)
            pos += sz
        await self._do_connection()

    async def _handle_decode_zrle(self, x, y, width, height):
        buffer = await self.reader.read(4)

        (compressed_bytes,) = unpack("!L", buffer)
        await self._handle_decode_zrl_edata(x, y, width, height, compressed_bytes)

    async def _handle_decode_zrl_edata(self, x, y, width, height, num):
        buffer = await self.reader.read(num)

        tx = x
        ty = y

        data = self._zlib_stream.decompress(buffer)
        it = iter(data)

        def cpixel(i):
            yield next(i)
            yield next(i)
            yield next(i)
            yield 0xff

        while True:
            try:
                subencoding = ord(next(it))
            except StopIteration:
                break

            tw = th = 64
            if x + width - tx < 64:
                tw = x + width - tx
            if y + height - ty < 64:
                th = y + height - ty

            pixels_in_tile = tw * th

            num_pixels = 0
            pixel_data = bytearray()
            palette_size = subencoding & 127
            if subencoding & 0x80:
                def do_rle(pixel):
                    run_length_next = ord(next(it))
                    run_length = run_length_next
                    while run_length_next == 255:
                        run_length_next = ord(next(it))
                        run_length += run_length_next
                    pixel_data.extend(pixel * (run_length + 1))
                    return run_length + 1

                if palette_size == 0:
                    while num_pixels < pixels_in_tile:
                        color = bytearray(cpixel(it))
                        num_pixels += do_rle(color)
                    if num_pixels != pixels_in_tile:
                        raise ValueError("too many pixels")
                else:
                    palette = [bytearray(cpixel(it)) for p in range(palette_size)]

                    while num_pixels < pixels_in_tile:
                        palette_index = ord(next(it))
                        if palette_index & 0x80:
                            palette_index &= 0x7F
                            num_pixels += do_rle(palette[palette_index])
                        else:
                            pixel_data.extend(palette[palette_index])
                            num_pixels += 1
                    if num_pixels != pixels_in_tile:
                        raise ValueError("too many pixels")

                await self.update_rectangle(tx, ty, tw, th, bytes(pixel_data))
            else:
                if palette_size == 0:
                    pixel_data = b''.join(bytes(cpixel(it)) for _ in range(pixels_in_tile))
                    await self.update_rectangle(tx, ty, tw, th, bytes(pixel_data))
                elif palette_size == 1:
                    color = bytearray(cpixel(it))
                    await self.fill_rectangle(tx, ty, tw, th, bytes(color))
                else:
                    if palette_size > 16:
                        raise ValueError("Palette of size {0} is not allowed".format(palette_size))

                    palette = [bytearray(cpixel(it)) for _ in range(palette_size)]
                    if palette_size == 2:
                        next_index = _zrle_next_bit(it, pixels_in_tile)
                    elif palette_size == 3 or palette_size == 4:
                        next_index = _zrle_next_dibit(it, pixels_in_tile)
                    else:
                        next_index = _zrle_next_nibble(it, pixels_in_tile)

                    for palette_index in next_index:
                        pixel_data.extend(palette[palette_index])
                    await self.update_rectangle(tx, ty, tw, th, bytes(pixel_data))

            tx = tx + 64
            if tx >= x + width:
                tx = x
                ty = ty + 64

        await self._do_connection()

    async def _handle_decode_psuedo_cursor(self, length, x, y, width, height):
        buffer = await self.reader.read(length)

        split = width * height * self.bypp
        image = buffer[:split]
        mask = buffer[split:]
        await self.update_cursor(x, y, width, height, image, mask)
        await self._do_connection()

    async def _handle_decode_desktop_size(self, width, height):
        await self.update_desktop_size(width, height)
        await self._do_connection()

    async def _handle_framebuffer_update(self):
        buffer = await self.reader.read(3)

        self.rectangles, = unpack("!xH", buffer)
        self.rectanglePos = []
        await self.begin_update()
        await self._do_connection()

    async def _handle_server_cut_text(self):
        buffer = await self.reader.read(7)
        (length,) = unpack("!xxxI", buffer)
        await self._handle_server_cut_text_value(length)

    async def _handle_server_cut_text_value(self, length):
        buffer = await self.reader.read(length)

        await self.copy_text(buffer)

    async def vnc_request_password(self):
        if self.password is None:
            print("need a password")
            await self.writer.close()
            return
        await self.send_password(self.password)

    async def vnc_connection_made(self):
        ...

    async def vnc_auth_failed(self, reason):
        print("Cannot connect %s" % reason)

    async def send_password(self, password):
        pw = (password + '\0' * 8)[:8]
        response = RFBDes(pw).encrypt(self._challenge)
        self.writer.write(response)

    async def begin_update(self):
        ...

    async def commit_update(self, positions):
        ...

    async def copy_rectangle(self, srcx, srcy, x, y, width, height):
        ...

    async def update_rectangle(self, x, y, width, height, buffer):
        ...

    async def fill_rectangle(self, x, y, width, height, color):
        await self.update_rectangle(x, y, width, height, color * width * height)

    async def update_cursor(self, x, y, width, height, image, mask):
        ...

    async def update_desktop_size(self, width, height):
        ...

    async def copy_text(self, text):
        ...

    async def bell(self):
        print('!!BELL!!')

    async def set_pixel_format(self, bpp=32, depth=24, bigendian=0, truecolor=1, redmax=255, greenmax=255, bluemax=255, redshift=0, greenshift=8, blueshift=16):
        pixformat = pack("!BBBBHHHBBBxxx", bpp, depth, bigendian, truecolor, redmax, greenmax, bluemax, redshift, greenshift, blueshift)
        self.writer.write(pack("!Bxxx16s", 0, pixformat))
        self.bpp, self.depth, self.bigendian, self.truecolor = bpp, depth, bigendian, truecolor
        self.redmax, self.greenmax, self.bluemax = redmax, greenmax, bluemax
        self.redshift, self.greenshift, self.blueshift = redshift, greenshift, blueshift
        self.bypp = self.bpp // 8

    async def set_encodings(self, list_of_encodings):
        self.writer.write(pack("!BxH", 2, len(list_of_encodings)))
        for encoding in list_of_encodings:
            self.writer.write(pack("!i", encoding))

    async def framebuffer_update_request(self, x=0, y=0, width=None, height=None, incremental=0):
        if width is None:
            width = self.width - x
        if height is None:
            height = self.height - y
        self.writer.write(pack("!BBHHHH", 3, incremental, x, y, width, height))

    async def key_event(self, key, down=1):
        self.writer.write(pack("!BBxxI", 4, down, key))

    async def pointer_event(self, x, y, buttonmask=0):
        self.writer.write(pack("!BBHH", 5, buttonmask, x, y))

    async def client_cut_text(self, message):
        self.writer.write(pack("!BxxxI", 6, len(message)) + message)


class RFBDes(DES):
    def set_key(self, key):
        newkey = []
        for ki in range(len(key)):
            bsrc = ord(key[ki])
            btgt = 0
            for i in range(8):
                if bsrc & (1 << i):
                    btgt = btgt | (1 << 7 - i)
            newkey.append(chr(btgt))
        super(RFBDes, self).set_key(newkey)



import warnings
from typing import Any

EPOCH_DEPRECATION_WARNING = (
    "The epoch parameter in `scheduler.step()` was not necessary and is being "
    "deprecated where possible. Please use `scheduler.step()` to step the "
    "scheduler. During the deprecation, if epoch is different from None, the "
    "closed form is used instead of the new chainable form, where available. "
)

SAVE_STATE_WARNING = "Please also save or load the state of the optimizer when saving or loading the scheduler."


def _calculate_output_dims(input_shape, kernel_shape, padding, stride):
    batch_size, _, input_height, input_width = input_shape
    out_filter_number, _, filter_height, filter_width = kernel_shape

    output_height = (input_height + padding[0] * 2 - filter_height) // stride + 1
    output_width = (input_width + padding[1] * 2 - filter_width) // stride + 1
    return batch_size, out_filter_number, output_height, output_width


def _forward_unimplemented(self, *inp: Any) -> None:
    r"""Defines the computation performed at every call.

    Should be overridden by all subclasses.

    .. note::
        Although the recipe for forward pass needs to be defined within
        this function, one should call the :class:`Module` instance afterwards
        instead of this since the former takes care of running the
        registered hooks while the latter silently ignores them.
    """
    raise NotImplementedError


def typename(o):
    if type(o) == 'Tensor':
        return o.dtype

    module = ''
    if hasattr(o, '__module__') and o.__module__ != 'builtins' and o.__module__ != '__builtin__' and o.__module__ is not None:
        module = o.__module__ + '.'

    if hasattr(o, '__qualname__'):
        class_name = o.__qualname__
    elif hasattr(o, '__name__'):
        class_name = o.__name__
    else:
        class_name = o.__class__.__name__

    return module + class_name


def indent(string, spaces):
    s = string.split('\n')
    if len(s) == 1:
        return string
    first = s.pop(0)
    s = [(spaces * ' ') + line for line in s]
    s = '\n'.join(s)
    s = first + '\n' + s
    return s


def get_enum(reduction):
    if reduction == 'none':
        ret = 0
    elif reduction == 'mean':
        ret = 1
    elif reduction == 'elementwise_mean':
        warnings.warn("reduction='elementwise_mean' is deprecated, please use reduction='mean' instead.")
        ret = 1
    elif reduction == 'sum':
        ret = 2
    else:
        ret = -1  # TODO: remove once JIT exceptions support control flow
        raise ValueError("{} is not a valid value for reduction".format(reduction))
    return ret


def legacy_get_string(size_average, reduce, emit_warning=True):
    warning = "size_average and reduce args will be deprecated, please use reduction='{}' instead."

    if size_average is None:
        size_average = True
    if reduce is None:
        reduce = True

    if size_average and reduce:
        ret = 'mean'
    elif reduce:
        ret = 'sum'
    else:
        ret = 'none'
    if emit_warning:
        warnings.warn(warning.format(ret))
    return ret


def legacy_get_enum(size_average, reduce, emit_warning=True):
    return get_enum(legacy_get_string(size_average, reduce, emit_warning))


class _RequiredParameter(object):
    """Singleton class representing a required parameter for an Optimizer."""

    def __repr__(self):
        return "<required parameter>"


required = _RequiredParameter()



from __future__ import print_function

import json
import smtplib
import time
from email.mime.text import MIMEText

from joatmon.assistant.service import BaseService
from joatmon.assistant.task import BaseTask
from joatmon.decorator.message import consumer
from joatmon.plugin.core import register
from joatmon.plugin.message.kafka import KafkaPlugin


class Task(BaseTask):
    def __init__(self, api, **kwargs):
        super(Task, self).__init__(api, **kwargs)

    @staticmethod
    def help():
        return {
            "name": "mail",
            "description": "a function for user to send a mail to someone",
            "parameters": {
                "type": "object",
                "properties": {
                    "subject": {
                        "type": "string",
                        "description": "subject of the mail"
                    },
                    "message": {
                        "type": "string",
                        "description": "content of the mail"
                    },
                    "receiver": {
                        "type": "string",
                        "description": "receiver of the mail"
                    }
                },
                "required": ["subject", "message", "receiver"]
            }
        }

    def run(self):
        subject = self.kwargs.get('subject', '') or self.api.input('what do you want the subject to be')
        message = self.kwargs.get('message', '') or self.api.input('what do you want the content to be')
        receiver = self.kwargs.get('receiver', '') or self.api.input('what do you want the receiver to be')

        contacts = json.loads(open('iva/iva.json', 'r').read()).get('contacts', [])
        contact = list(filter(lambda x: x['name'] == receiver, contacts))
        if len(contact) > 0:
            receiver = contact[0]['email']

        receivers = [receiver]

        config = json.loads(open('iva/iva.json', 'r').read())['config']['mail']

        text_subtype = 'plain'
        subject = subject

        server = config['server']
        write_port = config['write_port']
        address = config['address']
        password = config['password']

        msg = MIMEText(message, text_subtype)
        msg['Subject'] = subject
        msg['From'] = address

        conn = smtplib.SMTP(server, write_port)
        conn.login(address, password)
        conn.sendmail(address, receivers, msg.as_string())
        conn.quit()

        if not self.stop_event.is_set():
            self.stop_event.set()


class Service(BaseService):
    def __init__(self, api, **kwargs):
        super(Service, self).__init__(api, **kwargs)

        register(KafkaPlugin, 'kafka_plugin', 'localhost:9092')

        self.send_mail = consumer('kafka_plugin', 'mail')(self.send_mail)

    @staticmethod
    def help():
        return {}

    def send_mail(self, mail):
        config = json.loads(open('iva/iva.json', 'r').read())['config']['mail']

        text_subtype = mail['content_type']
        content = mail['content']
        subject = mail['subject']

        server = config['server']
        write_port = config['write_port']
        address = config['address']
        password = config['password']
        receivers = mail['receivers']

        msg = MIMEText(content, text_subtype)
        msg['Subject'] = subject
        msg['From'] = address

        conn = smtplib.SMTP(server, write_port)
        conn.login(address, password)
        conn.sendmail(address, receivers, msg.as_string())
        conn.quit()

    def run(self):
        while True:
            if self.event.is_set():
                break
            time.sleep(1)



import warnings

from joatmon.nn.core import (
    LRScheduler
)

__all__ = ['ExponentialLR']


class ExponentialLR(LRScheduler):
    """Decays the learning rate of each parameter group by gamma every epoch.
    When last_epoch=-1, sets initial lr as lr.

    Args:
        optimizer (Optimizer): Wrapped optimizer.
        gamma (float): Multiplicative factor of learning rate decay.
        last_epoch (int): The index of last epoch. Default: -1.
        verbose (bool): If ``True``, prints a message to stdout for
            each update. Default: ``False``.
    """

    def __init__(self, optimizer, gamma, last_epoch=-1, verbose=False):
        self.gamma = gamma
        super(ExponentialLR, self).__init__(optimizer, last_epoch, verbose)

    def get_lr(self):
        if not self._get_lr_called_within_step:
            warnings.warn("To get the last learning rate computed by the scheduler, please use `get_last_lr()`.", UserWarning)

        if self.last_epoch == 0:
            return self.base_lrs
        return [group['lr'] * self.gamma for group in self.optimizer.param_groups]

    def _get_closed_form_lr(self):
        return [base_lr * self.gamma ** self.last_epoch for base_lr in self.base_lrs]



import pygame

from joatmon.system.gui.widgets.widget import Element


class Label(Element):
    def __init__(self):
        super(Label, self).__init__()

        self._text = None
        self._text_color = (0, 0, 0)
        self._font = pygame.font.SysFont(None, 24)

    def text(self, text):
        self._text = text

        return self

    def style(self, style):
        super(Label, self).style(style)

        self._text_color = style.get('text_color', self._text_color)
        self._font = style.get('font', self._font)

    def draw(self, screen):
        super(Label, self).draw(screen)

        text_surface = self._font.render(self._text, True, self._text_color)
        text_rect = text_surface.get_rect(center=self._rect.center)
        screen.blit(text_surface, text_rect)



import json
import pickle
import random

import nltk
import numpy as np
import tensorflow as tf
from nltk.stem import WordNetLemmatizer

nltk.download('punkt', quiet=True)
nltk.download('wordnet', quiet=True)


class GenericAssistant:
    def __init__(self, intents):
        self.hist = None
        self.model = None
        self.classes = None
        self.words = None

        self.intents = intents

        if intents.endswith(".json"):
            self.load_json_intents(intents)

        self.lemmatizer = WordNetLemmatizer()

    def load_json_intents(self, intents):
        self.intents = json.loads(open(intents).read())

    def train_model(self):
        self.words = []
        self.classes = []
        documents = []
        ignore_letters = ['!', '?', ',', '.']

        for intent in self.intents['intents']:
            for pattern in intent['patterns']:
                word = nltk.word_tokenize(pattern)
                self.words.extend(word)
                documents.append((word, intent['name']))
                if intent['name'] not in self.classes:
                    self.classes.append(intent['name'])

        self.words = [self.lemmatizer.lemmatize(w.lower()) for w in self.words if w not in ignore_letters]
        self.words = sorted(list(set(self.words)))

        self.classes = sorted(list(set(self.classes)))

        training = []
        output_empty = [0] * len(self.classes)

        for doc in documents:
            bag = []
            word_patterns = doc[0]
            word_patterns = [self.lemmatizer.lemmatize(word.lower()) for word in word_patterns]
            for word in self.words:
                bag.append(1) if word in word_patterns else bag.append(0)

            output_row = list(output_empty)
            output_row[self.classes.index(doc[1])] = 1
            training.append([bag, output_row])

        random.shuffle(training)
        training = np.array(training)

        train_x = list(training[:, 0])
        train_y = list(training[:, 1])

        self.model = tf.keras.models.Sequential()
        self.model.add(tf.keras.layers.Dense(128, input_shape=(len(train_x[0]),), activation='relu'))
        self.model.add(tf.keras.layers.Dropout(0.5))
        self.model.add(tf.keras.layers.Dense(64, activation='relu'))
        self.model.add(tf.keras.layers.Dropout(0.5))
        self.model.add(tf.keras.layers.Dense(len(train_y[0]), activation='softmax'))

        sgd = tf.keras.optimizers.legacy.SGD(lr=0.01, decay=1e-6, momentum=0.9, nesterov=True)
        self.model.compile(loss='categorical_crossentropy', optimizer=sgd, metrics=['accuracy'])

        self.hist = self.model.fit(np.array(train_x), np.array(train_y), epochs=200, batch_size=5, verbose=0)

    def save_model(self, model_name=None):
        self.model.save(f"iva/weights/{model_name}/weights.h5", self.hist)
        pickle.dump(self.words, open(f'iva/weights/{model_name}/words.pkl', 'wb'))
        pickle.dump(self.classes, open(f'iva/weights/{model_name}/classes.pkl', 'wb'))

    def load_model(self, model_name=None):
        self.words = pickle.load(open(f'iva/weights/{model_name}/words.pkl', 'rb'))
        self.classes = pickle.load(open(f'iva/weights/{model_name}/classes.pkl', 'rb'))
        self.model = tf.keras.models.load_model(f'iva/weights/{model_name}.weights.h5')

    def _clean_up_sentence(self, sentence):
        sentence_words = nltk.word_tokenize(sentence)
        sentence_words = [self.lemmatizer.lemmatize(word.lower()) for word in sentence_words]
        return sentence_words

    def _bag_of_words(self, sentence, words):
        sentence_words = self._clean_up_sentence(sentence)
        bag = [0] * len(words)
        for s in sentence_words:
            for i, word in enumerate(words):
                if word == s:
                    bag[i] = 1
        return np.array(bag)

    def _predict_class(self, sentence):
        p = self._bag_of_words(sentence, self.words)
        res = self.model.predict(np.array([p]), verbose=0)[0]
        error_threshold = 0.1
        results = [[i, r] for i, r in enumerate(res) if r > error_threshold]

        results.sort(key=lambda x: x[1], reverse=True)
        return_list = []
        for r in results:
            return_list.append({'intent': self.classes[r[0]], 'probability': r[1]})
        return return_list

    def request(self, message):
        ints = self._predict_class(message)

        return ints[0]['intent'], ints[0]['probability']



from joatmon.assistant import (
    service,
    task
)
from joatmon.assistant.task import BaseTask


class Task(BaseTask):
    def __init__(self, api, **kwargs):
        super(Task, self).__init__(api, **kwargs)

    @staticmethod
    def help():
        return {
            "name": "create",
            "description": "a function for user to create service or task",
            "parameters": {
                "type": "object",
                "properties": {
                    "mode": {
                        "type": "string",
                        "enum": ["task", "service"]
                    }
                },
                "required": ["mode"]
            }
        }

    def run(self):
        if self.kwargs.get('mode', '') == 'task':
            task.create(self.api)
        if self.kwargs.get('mode', '') == 'service':
            service.create(self.api)

        if not self.stop_event.is_set():
            self.stop_event.set()



import os

import matplotlib.pyplot as plt
import numpy as np

plt.style.use('ggplot')


class CoreCallback(object):
    """
    Abstract base class for all implemented callback.

    Do not use this abstract base class directly but instead use one of the concrete callback implemented.

    To implement your own callback, you have to implement the following methods:

    - `on_action_begin`
    - `on_action_end`
    - `on_replay_begin`
    - `on_replay_end`
    - `on_episode_begin`
    - `on_episode_end`
    - `on_agent_begin`
    - `on_agent_end`
    """

    def __init__(self):
        super(CoreCallback, self).__init__()

    def on_agent_begin(self, *args, **kwargs):
        """
        Called at beginning of each agent play
        """
        pass

    def on_agent_end(self, *args, **kwargs):
        """
        Called at end of each agent play
        """
        pass

    def on_episode_begin(self, *args, **kwargs):
        """
        Called at beginning of each game episode
        """
        pass

    def on_episode_end(self, *args, **kwargs):
        """
        Called at end of each game episode
        """
        pass

    def on_action_begin(self, *args, **kwargs):
        """
        Called at beginning of each agent action
        """
        pass

    def on_action_end(self, *args, **kwargs):
        """
        Called at end of each agent action
        """
        pass

    def on_replay_begin(self, *args, **kwargs):
        """
        Called at beginning of each nn replay
        """
        pass

    def on_replay_end(self, *args, **kwargs):
        """
        Called at end of each nn replay
        """
        pass


class CallbackList(CoreCallback):
    def __init__(self, callbacks):
        super(CallbackList, self).__init__()

        self.callbacks = callbacks

    def on_action_begin(self, *args, **kwargs):
        for callback in self.callbacks:
            callback.on_action_begin(*args, **kwargs)

    def on_action_end(self, *args, **kwargs):
        for callback in self.callbacks:
            callback.on_action_end(*args, **kwargs)

    def on_agent_begin(self, *args, **kwargs):
        for callback in self.callbacks:
            callback.on_agent_begin(*args, **kwargs)

    def on_agent_end(self, *args, **kwargs):
        for callback in self.callbacks:
            callback.on_agent_end(*args, **kwargs)

    def on_episode_begin(self, *args, **kwargs):
        for callback in self.callbacks:
            callback.on_episode_begin(*args, **kwargs)

    def on_episode_end(self, *args, **kwargs):
        for callback in self.callbacks:
            callback.on_episode_end(*args, **kwargs)

    def on_replay_begin(self, *args, **kwargs):
        for callback in self.callbacks:
            callback.on_replay_begin(*args, **kwargs)

    def on_replay_end(self, *args, **kwargs):
        for callback in self.callbacks:
            callback.on_replay_end(*args, **kwargs)


class Loader(CoreCallback):
    def __init__(self, model, run_path, interval):
        super().__init__()

        self.network = model
        self.run_path = run_path
        self.interval = interval

    def on_agent_begin(self, *args, **kwargs):
        # weights path should be run path
        self.network.load(self.run_path)
        self.network.save(self.run_path)

    def on_agent_end(self, *args, **kwargs):
        self.network.save(self.run_path)

    def on_episode_end(self, *args, **kwargs):
        if 'episode_number' in kwargs:
            if kwargs['episode_number'] % self.interval == 0:
                self.network.save(self.run_path)


class Renderer(CoreCallback):
    def __init__(self, environment):
        super().__init__()

        self.environment = environment

    def on_action_end(self, *args, **kwargs):
        self.environment.render()

    def on_episode_begin(self, *args, **kwargs):
        self.environment.render()


class TrainLogger(CoreCallback):
    def __init__(self, run_path, interval):
        super().__init__()

        self.run_path = run_path
        self.interval = interval

        if not os.path.exists(run_path):
            os.makedirs(run_path)

        self.agent_data_path = self.run_path + 'train-agent-data.csv'
        self.network_data_path = self.run_path + 'train-nn-data.csv'
        self.episode_end_message_raw = '\repisode {:05d} ended in {:04d} actions, total reward: {:+08.2f}'

    def on_agent_begin(self, *args, **kwargs):
        if 'agent_headers' in kwargs:
            with open(self.agent_data_path, 'w') as file:
                file.write(','.join(kwargs['agent_headers']) + '\n')

        if 'network_headers' in kwargs:
            with open(self.network_data_path, 'w') as file:
                file.write(','.join(kwargs['network_headers']) + '\n')

    def on_episode_begin(self, *args, **kwargs):
        pass

    def on_episode_end(self, *args, **kwargs):
        if 'episode_number' in kwargs and 'action_number' in kwargs and 'episode_reward' in kwargs:
            end = '\n' if kwargs['episode_number'] % self.interval == 0 else ''
            print(self.episode_end_message_raw.format(kwargs['episode_number'], kwargs['action_number'], kwargs['episode_reward']), end=end)
            with open(self.agent_data_path, 'a') as file:
                file.write(','.join(list(map(str, [kwargs['episode_number'], kwargs['action_number'], kwargs['episode_reward']]))) + '\n')

    def on_replay_end(self, *args, **kwargs):
        if 'loss' in kwargs:
            if not isinstance(kwargs['loss'], (list, tuple)):
                kwargs['loss'] = [kwargs['loss']]

            with open(self.network_data_path, 'a') as file:
                file.write(','.join(list(map(str, kwargs['loss']))) + '\n')


class ValidationLogger(CoreCallback):
    def __init__(self, run_path, interval):
        super().__init__()

        self.run_path = run_path
        self.interval = interval

        if not os.path.exists(run_path):
            os.makedirs(run_path)

        self.agent_data_path = self.run_path + 'test-agent-data.csv'
        self.episode_end_message_raw = '\repisode {:05d} ended in {:04d} actions, total reward: {:+08.2f}'

    def on_agent_begin(self, *args, **kwargs):
        with open(self.agent_data_path, 'w') as file:
            file.write('episode_number,action_number,episode_reward\n')

    def on_episode_begin(self, *args, **kwargs):
        pass

    def on_episode_end(self, *args, **kwargs):
        if 'episode_number' in kwargs and 'action_number' in kwargs and 'episode_reward' in kwargs:
            end = '\n' if kwargs['episode_number'] % self.interval == 0 else ''
            print(self.episode_end_message_raw.format(kwargs['episode_number'], kwargs['action_number'], kwargs['episode_reward']), end=end)
            with open(self.agent_data_path, 'a') as file:
                file.write(','.join(list(map(str, [kwargs['episode_number'], kwargs['action_number'], kwargs['episode_reward']]))) + '\n')


class Visualizer(CoreCallback):
    def __init__(self, model, predicate=lambda x: True):
        super().__init__()

        self.model = model
        self.predicate = predicate

    def on_action_begin(self, *args, **kwargs):
        inputs = []
        if 'state' in kwargs:
            inputs.append(np.expand_dims(kwargs['state'], axis=0))

        if len(self.model.inputs) == 2 and 'action' in kwargs:
            inputs.append(np.expand_dims(kwargs['action'], axis=0))


class LivePlotter(CoreCallback):
    @staticmethod
    def live_plotter(y_num, size):
        plt.ion()
        fig = plt.figure(figsize=(13, 6))
        ax = fig.add_subplot(111)
        lines = ax.plot([i for i in range(size)], [0] * size, '-o', alpha=0.8)
        plt.ylabel('Y Label')
        plt.title('Title: {}'.format('1'))
        plt.show()

        def f(ys):
            lines[0].set_ydata(ys[0])
            # lines[0].set_xdata(x)  # if this is done, need to set xlim as well
            if np.min(ys[0]) < lines[0].axes.get_ylim()[0] or np.max(ys[0]) > lines[0].axes.get_ylim()[1]:
                plt.ylim([np.min(ys[0]) - np.std(ys[0]), np.max(ys[0]) + np.std(ys[0])])
            plt.pause(.1)

        return f



import warnings

from elasticsearch import Elasticsearch

from joatmon.plugin.database.core import DatabasePlugin


class ElasticDatabase(DatabasePlugin):
    DATABASES = set()
    CREATED_COLLECTIONS = set()
    UPDATED_COLLECTIONS = set()

    def __init__(self, uri):
        self.client = Elasticsearch(uri)

    async def insert(self, document, *docs):
        for doc in docs:
            if document.__metaclass__.structured:
                warnings.warn(f'document validation will be ignored')

            self.client.index(index=document.__metaclass__.__collection__, document=dict(**doc))

    async def read(self, document, query):
        ...

    async def update(self, document, query, update):
        ...

    async def delete(self, document, query):
        ...

    async def start(self):
        ...

    async def commit(self):
        ...

    async def abort(self):
        ...

    async def end(self):
        ...



import importlib.util
import inspect
import json
import os

from joatmon.assistant.task import BaseTask


class Task(BaseTask):
    def __init__(self, api, **kwargs):
        super(Task, self).__init__(api, **kwargs)

    @staticmethod
    def help():
        return {
            "name": "help",
            "description": "a function for user to learn about a function",
            "parameters": {
                "type": "object",
                "properties": {
                    "function": {
                        "type": "string",
                        "description": "name of the function that the user want to learn about"
                    }
                },
                "required": ["name"]
            }
        }

    def run(self):
        settings = json.loads(open('iva/iva.json', 'r').read())
        for scripts in settings.get('scripts', []):
            if os.path.isabs(scripts) and os.path.exists(scripts):
                for module in list(filter(lambda x: '__' not in x, map(lambda x: x.replace('.py', ''), os.listdir(scripts.replace('.', '/'))))):
                    spec = importlib.util.spec_from_file_location(module, os.path.join(scripts, f'{module}.py'))
                    action_module = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(action_module)

                    task = getattr(action_module, 'Task', None)
                    if task is None:
                        continue
                    task.help(self)
            else:
                _module = __import__(scripts, fromlist=[''])

                for module in inspect.getmembers(_module, predicate=inspect.ismodule):
                    action_module = getattr(_module, module[0])

                    task = getattr(action_module, 'Task', None)
                    if task is None:
                        continue
                    task.help(self)

        if not self.stop_event.is_set():
            self.stop_event.set()



def binary_search(iterable, value, item_getter):
    low = 0
    high = len(iterable) - 1

    while low <= high:
        mid = low + (high - low) // 2

        if item_getter(iterable[mid]) == value:
            return iterable[mid]
        elif item_getter(iterable[mid]) < 0:
            low = mid + 1
        else:
            high = mid - 1


def binary_search_helper(iterable, value, item_getter, low, high):
    if high >= low:
        mid = low + (high - low) // 2

        if item_getter(iterable[mid]) == value:
            return iterable[mid]
        elif item_getter(iterable[mid]) > 0:
            return binary_search_helper(iterable, value, item_getter, low, mid - 1)
        else:
            return binary_search_helper(iterable, value, item_getter, mid + 1, high)


def binary_search_r(iterable, value, item_getter):
    return binary_search_helper(iterable, value, item_getter, 0, len(iterable) - 1)



import pyotp

from joatmon.plugin.otp.core import OTP


class TOTP(OTP):
    async def get_qr(self, secret, name, issuer):
        return pyotp.totp.TOTP(secret).provisioning_uri(name=name, issuer_name=issuer)

    async def verify(self, secret, otp):
        return pyotp.TOTP(secret).verify(otp)



from __future__ import print_function

from joatmon.assistant.task import BaseTask
from joatmon.search import summary


class Task(BaseTask):
    def __init__(self, api, **kwargs):
        super(Task, self).__init__(api, **kwargs)

    @staticmethod
    def help():
        return {
            "name": "search",
            "description": "a function for user to search an expression",
            "parameters": {
                "type": "object",
                "properties": {
                    "message": {
                        "type": "string",
                        "description": "message to search for"
                    }
                },
                "required": ["message"]
            }
        }

    def run(self):
        message = self.kwargs.get('message', '') or self.api.input('what do you want the message to be')

        result = summary(message)
        self.api.output(result)

        if not self.stop_event.is_set():
            self.stop_event.set()
